"""
In-Game Mail system

Evennia Contribution - grungies1138 2016

A simple Brandymail style @mail system that uses the Msg class from Evennia
Core. It has two Commands, both of which can be used on their own:

   - CmdMail - this should sit on the Account cmdset and makes the `mail` command
    available both IC and OOC. Mails will always go to Accounts (other players).
   - CmdMailCharacter - this should sit on the Character cmdset and makes the `mail`
    command ONLY available when puppeting a character. Mails will be sent to other
    Characters only and will not be available when OOC.
   - If adding *both* commands to their respective cmdsets, you'll get two separate
    IC and OOC mailing systems, with different lists of mail for IC and OOC modes.

Installation:

Install one or both of the following (see above):

- CmdMail (IC + OOC mail, sent between players)

    # mygame/commands/default_cmds.py

    from evennia.contrib.game_systems import mail

    # in AccountCmdSet.at_cmdset_creation:
        self.add(mail.CmdMail())

- CmdMailCharacter (optional, IC only mail, sent between characters)

    # mygame/commands/default_cmds.py

    from evennia.contrib.game_systems import mail

    # in CharacterCmdSet.at_cmdset_creation:
        self.add(mail.CmdMailCharacter())

Once installed, use `help mail` in game for help with the mail command. Use
ic/ooc to switch in and out of IC/OOC modes.

"""

import re

from evennia import AccountDB, ObjectDB, default_cmds
from evennia.comms.models import Msg
from evennia.utils import create, datetime_format, evtable, inherits_from, make_iter

_HEAD_CHAR = "|015-|n"
_SUB_HEAD_CHAR = "-"
_WIDTH = 78


class CmdMail(default_cmds.MuxAccountCommand):
    """
    Communicate with others by sending mail.

    Usage:
      @mail       - Displays all the mail an account has in their mailbox
      @mail <#>   - Displays a specific message
      @mail <accounts>=<subject>/<message>
              - Sends a message to the comma separated list of accounts.
      @mail/delete <#> - Deletes a specific message
      @mail/forward <account list>=<#>[/<Message>]
              - Forwards an existing message to the specified list of accounts,
                original message is delivered with optional Message prepended.
      @mail/reply <#>=<message>
              - Replies to a message #. Prepends message to the original
                message text.
    Switches:
      delete  - deletes a message
      forward - forward a received message to another object with an optional message attached.
      reply   - Replies to a received message, appending the original message to the bottom.
    Examples:
      @mail 2
      @mail Griatch=New mail/Hey man, I am sending you a message!
      @mail/delete 6
      @mail/forward feend78 Griatch=4/You guys should read this.
      @mail/reply 9=Thanks for the info!

    """

    key = "@mail"
    aliases = ["mail"]
    lock = "cmd:all()"
    help_category = "General"

    def parse(self):
        """
        Add convenience check to know if caller is an Account or not since this cmd
        will be able to add to either Object- or Account level.

        """
        super().parse()
        self.caller_is_account = bool(
            inherits_from(self.caller, "evennia.accounts.accounts.DefaultAccount")
        )

    def search_targets(self, namelist):
        """
        Search a list of targets of the same type as caller.

        Args:
            caller (Object or Account): The type of object to search.
            namelist (list): List of strings for objects to search for.

        Returns:
            targetlist (Queryset): Any target matches.

        """
        nameregex = r"|".join(r"^%s$" % re.escape(name) for name in make_iter(namelist))
        if self.caller_is_account:
            matches = AccountDB.objects.filter(username__iregex=nameregex)
        else:
            matches = ObjectDB.objects.filter(db_key__iregex=nameregex)
        return matches

    def get_all_mail(self):
        """
        Returns a list of all the messages where the caller is a recipient. These
            are all messages tagged with tags of the `mail` category.

        Returns:
            messages (QuerySet): Matching Msg objects.

        """
        if self.caller_is_account:
            return Msg.objects.get_by_tag(category="mail").filter(db_receivers_accounts=self.caller)
        else:
            return Msg.objects.get_by_tag(category="mail").filter(db_receivers_objects=self.caller)

    def send_mail(self, recipients, subject, message, caller):
        """
        Function for sending new mail.  Also useful for sending notifications
        from objects or systems.

        Args:
            recipients (list): list of Account or Character objects to receive
                the newly created mails.
            subject (str): The header or subject of the message to be delivered.
            message (str): The body of the message being sent.
            caller (obj): The object (or Account or Character) that is sending the message.

        """
        for recipient in recipients:
            recipient.msg("You have received a new @mail from %s" % caller)
            new_message = create.create_message(
                self.caller, message, receivers=recipient, header=subject
            )
            new_message.tags.add("new", category="mail")

        if recipients:
            caller.msg("You sent your message.")
            return
        else:
            caller.msg("No valid target(s) found. Cannot send message.")
            return

    def func(self):
        """
        Do the main command functionality
        """

        subject = ""
        body = ""

        if self.switches or self.args:
            if "delete" in self.switches or "del" in self.switches:
                try:
                    if not self.lhs:
                        self.caller.msg("No Message ID given. Unable to delete.")
                        return
                    else:
                        all_mail = self.get_all_mail()
                        mind_max = max(0, all_mail.count() - 1)
                        mind = max(0, min(mind_max, int(self.lhs) - 1))
                        if all_mail[mind]:
                            mail = all_mail[mind]
                            question = "Delete message {} ({}) [Y]/N?".format(mind + 1, mail.header)
                            ret = yield (question)
                            # handle not ret, it will be None during unit testing
                            if not ret or ret.strip().upper() not in ("N", "No"):
                                all_mail[mind].delete()
                                self.caller.msg("Message %s deleted" % (mind + 1,))
                            else:
                                self.caller.msg("Message not deleted.")
                        else:
                            raise IndexError
                except IndexError:
                    self.caller.msg("That message does not exist.")
                except ValueError:
                    self.caller.msg("Usage: @mail/delete <message ID>")
            elif "forward" in self.switches or "fwd" in self.switches:
                try:
                    if not self.rhs:
                        self.caller.msg(
                            "Cannot forward a message without a target list. " "Please try again."
                        )
                        return
                    elif not self.lhs:
                        self.caller.msg("You must define a message to forward.")
                        return
                    else:
                        all_mail = self.get_all_mail()
                        mind_max = max(0, all_mail.count() - 1)
                        if "/" in self.rhs:
                            message_number, message = self.rhs.split("/", 1)
                            mind = max(0, min(mind_max, int(message_number) - 1))

                            if all_mail[mind]:
                                old_message = all_mail[mind]

                                self.send_mail(
                                    self.search_targets(self.lhslist),
                                    "FWD: " + old_message.header,
                                    message
                                    + "\n---- Original Message ----\n"
                                    + old_message.message,
                                    self.caller,
                                )
                                self.caller.msg("Message forwarded.")
                            else:
                                raise IndexError
                        else:
                            mind = max(0, min(mind_max, int(self.rhs) - 1))
                            if all_mail[mind]:
                                old_message = all_mail[mind]
                                self.send_mail(
                                    self.search_targets(self.lhslist),
                                    "FWD: " + old_message.header,
                                    "\n---- Original Message ----\n" + old_message.message,
                                    self.caller,
                                )
                                self.caller.msg("Message forwarded.")
                                old_message.tags.remove("new", category="mail")
                                old_message.tags.add("fwd", category="mail")
                            else:
                                raise IndexError
                except IndexError:
                    self.caller.msg("Message does not exist.")
                except ValueError:
                    self.caller.msg("Usage: @mail/forward <account list>=<#>[/<Message>]")
            elif "reply" in self.switches or "rep" in self.switches:
                try:
                    if not self.rhs:
                        self.caller.msg("You must define a message to reply to.")
                        return
                    elif not self.lhs:
                        self.caller.msg("You must supply a reply message")
                        return
                    else:
                        all_mail = self.get_all_mail()
                        mind_max = max(0, all_mail.count() - 1)
                        mind = max(0, min(mind_max, int(self.lhs) - 1))
                        if all_mail[mind]:
                            old_message = all_mail[mind]
                            self.send_mail(
                                old_message.senders,
                                "RE: " + old_message.header,
                                self.rhs + "\n---- Original Message ----\n" + old_message.message,
                                self.caller,
                            )
                            old_message.tags.remove("new", category="mail")
                            old_message.tags.add("-", category="mail")
                            return
                        else:
                            raise IndexError
                except IndexError:
                    self.caller.msg("Message does not exist.")
                except ValueError:
                    self.caller.msg("Usage: @mail/reply <#>=<message>")
            else:
                # normal send
                if self.rhs:
                    if "/" in self.rhs:
                        subject, body = self.rhs.split("/", 1)
                    else:
                        body = self.rhs
                    self.send_mail(self.search_targets(self.lhslist), subject, body, self.caller)
                else:
                    all_mail = self.get_all_mail()
                    mind_max = max(0, all_mail.count() - 1)
                    try:
                        mind = max(0, min(mind_max, int(self.lhs) - 1))
                        message = all_mail[mind]
                    except (ValueError, IndexError):
                        self.caller.msg("'%s' is not a valid mail id." % self.lhs)
                        return

                    messageForm = []
                    if message:
                        messageForm.append(_HEAD_CHAR * _WIDTH)
                        messageForm.append(
                            "|wFrom:|n %s" % (message.senders[0].get_display_name(self.caller))
                        )
                        # note that we cannot use %-d format here since Windows does not support it
                        day = message.db_date_created.day
                        messageForm.append(
                            "|wSent:|n %s"
                            % message.db_date_created.strftime(f"%b {day}, %Y - %H:%M:%S")
                        )
                        messageForm.append("|wSubject:|n %s" % message.header)
                        messageForm.append(_SUB_HEAD_CHAR * _WIDTH)
                        messageForm.append(message.message)
                        messageForm.append(_HEAD_CHAR * _WIDTH)
                    self.caller.msg("\n".join(messageForm))
                    message.tags.remove("new", category="mail")
                    message.tags.add("-", category="mail")

        else:
            # list messages
            messages = self.get_all_mail()

            if messages:
                table = evtable.EvTable(
                    "|wID|n",
                    "|wFrom|n",
                    "|wSubject|n",
                    "|wArrived|n",
                    "",
                    table=None,
                    border="header",
                    header_line_char=_SUB_HEAD_CHAR,
                    width=_WIDTH,
                )
                index = 1
                for message in messages:
                    status = str(message.db_tags.last().db_key.upper())
                    if status == "NEW":
                        status = "|gNEW|n"

                    table.add_row(
                        index,
                        message.senders[0].get_display_name(self.caller),
                        message.header,
                        datetime_format(message.db_date_created),
                        status,
                    )
                    index += 1

                table.reformat_column(0, width=6)
                table.reformat_column(1, width=18)
                table.reformat_column(2, width=34)
                table.reformat_column(3, width=13)
                table.reformat_column(4, width=7)

                self.caller.msg(_HEAD_CHAR * _WIDTH)
                self.caller.msg(str(table))
                self.caller.msg(_HEAD_CHAR * _WIDTH)
            else:
                self.caller.msg("There are no messages in your inbox.")


# character - level version of the command


class CmdMailCharacter(CmdMail):
    account_caller = False



"""
Resources for indexing help entries and for splitting help entries into
sub-categories.

This is used primarily by the default `help` command.

"""
import re

from django.conf import settings

# these are words that Lunr normally ignores but which we want to find
# since we use them (e.g. as command names).
# Lunr's default ignore-word list is found here:
# https://github.com/yeraydiazdiaz/lunr.py/blob/master/lunr/stop_word_filter.py
_LUNR_STOP_WORD_FILTER_EXCEPTIONS = [
    "about",
    "might",
    "get",
    "who",
    "say",
    "where",
] + settings.LUNR_STOP_WORD_FILTER_EXCEPTIONS


_LUNR = None
_LUNR_EXCEPTION = None

_LUNR_GET_BUILDER = None
_LUNR_BUILDER_PIPELINE = None

_RE_HELP_SUBTOPICS_START = re.compile(r"^\s*?#\s*?subtopics\s*?$", re.I + re.M)
_RE_HELP_SUBTOPIC_SPLIT = re.compile(r"^\s*?(\#{2,6}\s*?\w+?[a-z0-9 \-\?!,\.]*?)$", re.M + re.I)
_RE_HELP_SUBTOPIC_PARSE = re.compile(r"^(?P<nesting>\#{2,6})\s*?(?P<name>.*?)$", re.I + re.M)

MAX_SUBTOPIC_NESTING = 5


def help_search_with_index(query, candidate_entries, suggestion_maxnum=5, fields=None):
    """
    Lunr-powered fast index search and suggestion wrapper. See https://lunrjs.com/.

    Args:
        query (str): The query to search for.
        candidate_entries (list): This is the body of possible entities to search. Each
            must have a property `.search_index_entry` that returns a dict with all
            keys in the `fields` arg.
        suggestion_maxnum (int): How many matches to allow at most in a multi-match.
        fields (list, optional): A list of Lunr field mappings
            ``{"field_name": str, "boost": int}``. See the Lunr documentation
            for more details. The field name must exist in the dicts returned
            by `.search_index_entry` of the candidates. If not given, a default setup
            is used, prefering keys > aliases > category > tags.
    Returns:
        tuple: A tuple (matches, suggestions), each a list, where the `suggestion_maxnum` limits
            how many suggestions are included.

    """
    global _LUNR, _LUNR_EXCEPTION, _LUNR_BUILDER_PIPELINE, _LUNR_GET_BUILDER
    if not _LUNR:
        # we have to delay-load lunr because it messes with logging if it's imported
        # before twisted's logging has been set up
        from lunr import get_default_builder as _LUNR_GET_BUILDER
        from lunr import lunr as _LUNR
        from lunr import stop_word_filter
        from lunr.exceptions import QueryParseError as _LUNR_EXCEPTION
        from lunr.stemmer import stemmer

        # from lunr.trimmer import trimmer
        # pre-create a lunr index-builder pipeline where we've removed some of
        # the stop-words from the default in lunr.

        stop_words = stop_word_filter.WORDS

        for ignore_word in _LUNR_STOP_WORD_FILTER_EXCEPTIONS:
            try:
                stop_words.remove(ignore_word)
            except ValueError:
                pass

        custom_stop_words_filter = stop_word_filter.generate_stop_word_filter(stop_words)
        # _LUNR_BUILDER_PIPELINE = (trimmer, custom_stop_words_filter, stemmer)
        _LUNR_BUILDER_PIPELINE = (custom_stop_words_filter, stemmer)

    indx = [cnd.search_index_entry for cnd in candidate_entries]
    mapping = {indx[ix]["key"]: cand for ix, cand in enumerate(candidate_entries)}

    if not fields:
        fields = [
            {"field_name": "key", "boost": 10},
            {"field_name": "aliases", "boost": 7},
            {"field_name": "category", "boost": 6},
            {"field_name": "tags", "boost": 5},
        ]

    # build the search index
    builder = _LUNR_GET_BUILDER()
    builder.pipeline.reset()
    builder.pipeline.add(*_LUNR_BUILDER_PIPELINE)

    search_index = _LUNR(ref="key", fields=fields, documents=indx, builder=builder)

    try:
        matches = search_index.search(query)[:suggestion_maxnum]
    except _LUNR_EXCEPTION:
        # this is a user-input problem
        matches = []

    # matches (objs), suggestions (strs)
    return (
        [mapping[match["ref"]] for match in matches],
        [str(match["ref"]) for match in matches],  # + f" (score {match['score']})")   # good debug
    )


def parse_entry_for_subcategories(entry):
    """
    Parse a command docstring for special sub-category blocks:

    Args:
        entry (str): A help entry to parse

    Returns:
        dict: The dict is a mapping that splits the entry into subcategories. This
            will always hold a key `None` for the main help entry and
            zero or more keys holding the subcategories. Each is itself
            a dict with a key `None` for the main text of that subcategory
            followed by any sub-sub-categories down to a max-depth of 5.

    Example:
    ::

        '''
        Main topic text

        # SUBTOPICS

        ## foo

        A subcategory of the main entry, accessible as `help topic foo`
        (or using /, like `help topic/foo`)

        ## bar

        Another subcategory, accessed as `help topic bar`
        (or `help topic/bar`)

        ### moo

        A subcategory of bar, accessed as `help bar moo`
        (or `help bar/moo`)

        #### dum

        A subcategory of moo, accessed `help bar moo dum`
        (or `help bar/moo/dum`)

        '''

    This will result in this returned entry structure:
    ::

        {
           None: "Main topic text":
           "foo": {
                None: "main topic/foo text"
           },
           "bar": {
                None: "Main topic/bar text",
                "moo": {
                    None: "topic/bar/moo text"
                    "dum": {
                        None: "topic/bar/moo/dum text"
                    }
                }
           }
        }

    """
    topic, *subtopics = _RE_HELP_SUBTOPICS_START.split(entry, maxsplit=1)
    structure = {None: topic.strip("\n")}

    if subtopics:
        subtopics = subtopics[0]
    else:
        return structure

    keypath = []
    current_nesting = 0
    subtopic = None

    # from evennia import set_trace;set_trace()
    for part in _RE_HELP_SUBTOPIC_SPLIT.split(subtopics.strip()):
        subtopic_match = _RE_HELP_SUBTOPIC_PARSE.match(part.strip())
        if subtopic_match:
            # a new sub(-sub..) category starts.
            mdict = subtopic_match.groupdict()
            subtopic = mdict["name"].lower().strip()
            new_nesting = len(mdict["nesting"]) - 1

            if new_nesting > MAX_SUBTOPIC_NESTING:
                raise RuntimeError(
                    f"Can have max {MAX_SUBTOPIC_NESTING} levels of nested help subtopics."
                )

            nestdiff = new_nesting - current_nesting
            if nestdiff < 0:
                # jumping back up in nesting
                for _ in range(abs(nestdiff) + 1):
                    try:
                        keypath.pop()
                    except IndexError:
                        pass
            elif nestdiff == 0:
                # don't add a deeper nesting but replace the current
                try:
                    keypath.pop()
                except IndexError:
                    pass
            keypath.append(subtopic)
            current_nesting = new_nesting
        else:
            # an entry belonging to a subtopic - find the nested location
            dct = structure
            if not keypath and subtopic is not None:
                structure[subtopic] = part
            else:
                for key in keypath:
                    if key in dct:
                        dct = dct[key]
                    else:
                        dct[key] = {None: part}
    return structure



"""
Tutorial - talking NPC tests.

"""
from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.utils.create import create_object

from . import talking_npc


class TestTalkingNPC(BaseEvenniaCommandTest):
    def test_talkingnpc(self):
        npc = create_object(talking_npc.TalkingNPC, key="npctalker", location=self.room1)
        self.call(talking_npc.CmdTalk(), "", "(You walk up and talk to Char.)")
        npc.delete()



"""
Tests for the bodyfunctions.

"""
from mock import Mock, patch

from evennia.utils.test_resources import BaseEvenniaTest

from .bodyfunctions import BodyFunctions


@patch("evennia.contrib.tutorials.bodyfunctions.bodyfunctions.random")
class TestBodyFunctions(BaseEvenniaTest):
    script_typeclass = BodyFunctions

    def setUp(self):
        super().setUp()
        self.script.obj = self.char1

    def tearDown(self):
        super().tearDown()
        # if we forget to stop the script, DirtyReactorAggregateError will be raised
        self.script.stop()

    def test_at_repeat(self, mock_random):
        """test that no message will be sent when below the 66% threshold"""
        mock_random.random = Mock(return_value=0.5)
        old_func = self.script.send_random_message
        self.script.send_random_message = Mock()
        self.script.at_repeat()
        self.script.send_random_message.assert_not_called()
        # test that random message will be sent
        mock_random.random = Mock(return_value=0.7)
        self.script.at_repeat()
        self.script.send_random_message.assert_called()
        self.script.send_random_message = old_func

    def test_send_random_message(self, mock_random):
        """Test that correct message is sent for each random value"""
        old_func = self.char1.msg
        self.char1.msg = Mock()
        # test each of the values
        mock_random.random = Mock(return_value=0.05)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You tap your foot, looking around.")
        mock_random.random = Mock(return_value=0.15)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You have an itch. Hard to reach too.")
        mock_random.random = Mock(return_value=0.25)
        self.script.send_random_message()
        self.char1.msg.assert_called_with(
            "You think you hear someone behind you. ... " "but when you look there's noone there."
        )
        mock_random.random = Mock(return_value=0.35)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You inspect your fingernails. Nothing to report.")
        mock_random.random = Mock(return_value=0.45)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You cough discreetly into your hand.")
        mock_random.random = Mock(return_value=0.55)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You scratch your head, looking around.")
        mock_random.random = Mock(return_value=0.65)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You blink, forgetting what it was you were going to do.")
        mock_random.random = Mock(return_value=0.75)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You feel lonely all of a sudden.")
        mock_random.random = Mock(return_value=0.85)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You get a great idea. Of course you won't tell anyone.")
        mock_random.random = Mock(return_value=0.95)
        self.script.send_random_message()
        self.char1.msg.assert_called_with("You suddenly realize how much you love Evennia!")
        self.char1.msg = old_func



"""Tests for text2bbcode """

import mock
from django.test import TestCase

from evennia.contrib.base_systems.godotwebsocket import text2bbcode
from evennia.utils import ansi


class TestText2Bbcode(TestCase):
    def test_format_styles(self):
        parser = text2bbcode.BBCODE_PARSER
        self.assertEqual("foo", parser.format_styles("foo"))
        self.assertEqual(
            "[color=#800000]red[/color]foo",
            parser.format_styles(
                ansi.ANSI_UNHILITE + ansi.ANSI_RED + "red" + ansi.ANSI_NORMAL + "foo"
            ),
        )
        self.assertEqual(
            "[bgcolor=#800000]red[/bgcolor]foo",
            parser.format_styles(ansi.ANSI_BACK_RED + "red" + ansi.ANSI_NORMAL + "foo"),
        )
        self.assertEqual(
            "[bgcolor=#800000][color=#008000]red[/color][/bgcolor]foo",
            parser.format_styles(
                ansi.ANSI_BACK_RED
                + ansi.ANSI_UNHILITE
                + ansi.ANSI_GREEN
                + "red"
                + ansi.ANSI_NORMAL
                + "foo"
            ),
        )
        self.assertEqual(
            "a [u]red[/u]foo",
            parser.format_styles("a " + ansi.ANSI_UNDERLINE + "red" + ansi.ANSI_NORMAL + "foo"),
        )
        self.assertEqual(
            "a [blink]red[/blink]foo",
            parser.format_styles("a " + ansi.ANSI_BLINK + "red" + ansi.ANSI_NORMAL + "foo"),
        )
        self.assertEqual(
            "a [bgcolor=#c0c0c0][color=#000000]red[/color][/bgcolor]foo",
            parser.format_styles("a " + ansi.ANSI_INVERSE + "red" + ansi.ANSI_NORMAL + "foo"),
        )

    def test_convert_urls(self):
        parser = text2bbcode.BBCODE_PARSER
        self.assertEqual("foo", parser.convert_urls("foo"))
        self.assertEqual(
            "a [url=http://redfoo]http://redfoo[/url] runs",
            parser.convert_urls("a http://redfoo runs"),
        )

    def test_sub_mxp_links(self):
        parser = text2bbcode.BBCODE_PARSER
        mocked_match = mock.Mock()
        mocked_match.groups.return_value = ["cmd", "text"]

        self.assertEqual("[mxp=send cmd=cmd]text[/mxp]", parser.sub_mxp_links(mocked_match))

    def test_sub_text(self):
        parser = text2bbcode.BBCODE_PARSER

        mocked_match = mock.Mock()

        mocked_match.groupdict.return_value = {"lineend": "foo"}
        self.assertEqual("\n", parser.sub_text(mocked_match))

    def test_parse_bbcode(self):
        self.assertEqual("foo", text2bbcode.parse_to_bbcode("foo"))
        self.maxDiff = None
        self.assertEqual(
            text2bbcode.parse_to_bbcode("|^|[CHello|n|u|rW|go|yr|bl|md|c!|[G!"),
            "[blink][bgcolor=#008080]Hello[/bgcolor][/blink]"
            "[u][color=#ff0000]W[/color][/u]"
            "[u][color=#00ff00]o[/color][/u]"
            "[u][color=#ffff00]r[/color][/u]"
            "[u][color=#0000ff]l[/color][/u]"
            "[u][color=#ff00ff]d[/color][/u]"
            "[u][color=#00ffff]![/color][/u]"
            "[u][bgcolor=#008000][color=#00ffff]![/color][/bgcolor][/u]",
        )



"""
Buffs - Tegiminis 2022

A buff is a timed object, attached to a game entity, that modifies values, triggers 
code, or both. It is a common design pattern in RPGs, particularly action games.

This contrib gives you a buff handler to apply to your objects, a buff class to extend them,
a sample property class to show how to automatically check modifiers, some sample buffs to learn from,
and a command which applies buffs.

## Installation
Assign the handler to a property on the object, like so.

```python
@lazy_property
def buffs(self) -> BuffHandler:
    return BuffHandler(self)```

## Using the Handler

To make use of the handler, you will need:

- Some buffs to add. You can create these by extending the `BaseBuff` class from this module. You can see some examples in `samplebuffs.py`.
- A way to add buffs to the handler. You can see a basic example of this in the `CmdBuff` command in this module.

### Applying a Buff

Call the handler `add(BuffClass)` method. This requires a class reference, and also contains a number of 
optional arguments to customize the buff's duration, stacks, and so on.

```python
self.buffs.add(StrengthBuff)    # A single stack of StrengthBuff with normal duration
self.buffs.add(DexBuff, stacks=3, duration=60)  # Three stacks of DexBuff, with a duration of 60 seconds
self.buffs.add(ReflectBuff, to_cache={'reflect': 0.5})  # A single stack of ReflectBuff, with an extra cache value
```

### Modify

Call the handler `check(value, stat)` method wherever you want to see the modified value. 
This will return the value, modified by and relevant buffs on the handler's owner (identified by 
the `stat` string). For example:

```python
# The method we call to damage ourselves
def take_damage(self, source, damage):
    _damage = self.buffs.check(damage, 'taken_damage')
    self.db.health -= _damage
```

### Trigger

Call the handler `trigger(triggerstring)` method wherever you want an event call. This 
will call the `at_trigger` hook method on all buffs with the relevant trigger.

```python
def Detonate(BaseBuff):
    ...
    triggers = ['take_damage']
    def at_trigger(self, trigger, *args, **kwargs)
        self.owner.take_damage(100)
        self.remove()

def Character(Character):
    ...
    def take_damage(self, source, damage):
        self.buffs.trigger('take_damage')
        self.db.health -= _damage
```

### Tick

Ticking a buff happens automatically once applied, as long as the buff's `tickrate` is more than 0.

```python
def Poison(BaseBuff):
    ...
    tickrate = 5
    def at_tick(self, initial=True, *args, **kwargs):
        _dmg = self.dmg * self.stacks
        if not initial:
            self.owner.location.msg_contents(
                "Poison courses through {actor}'s body, dealing {damage} damage.".format(
                    actor=self.owner.named, damage=_dmg
                )
            )
```

## Buffs

A buff is a class which contains a bunch of immutable data about itself - such as tickrate, triggers, refresh rules, and
so on - and which merges mutable data in from the cache when called.

Buffs are always instanced when they are called for a method. To access a buff's properties and methods, you should do so through
this instance, rather than directly manipulating the buff cache on the object. You can modify a buff's cache through various handler
methods instead.

You can see all the features of the `BaseBuff` class below, or browse `samplebuffs.py` to see how to create some common buffs. Buffs have
many attributes and hook methods you can overload to create complex, interrelated buffs.

"""
import time
from random import random

from evennia import Command
from evennia.server import signals
from evennia.typeclasses.attributes import AttributeProperty
from evennia.utils import search, utils


class BaseBuff:
    key = "template"  # The buff's unique key. Will be used as the buff's key in the handler
    name = "Template"  # The buff's name. Used for user messaging
    flavor = "Template"  # The buff's flavor text. Used for user messaging
    visible = True  # If the buff is considered "visible" to the "view" method

    triggers = []  # The effect's trigger strings, used for functions.

    handler = None
    start = 0

    duration = -1  # Default buff duration; -1 for permanent, 0 for "instant", >0 normal
    playtime = False  # Does this buff autopause when owning object is unpuppeted?

    refresh = True  # Does the buff refresh its timer on application?
    unique = True  # Does the buff overwrite existing buffs with the same key on the same target?
    maxstacks = 1  # The maximum number of stacks the buff can have. If >1, this buff will stack.
    stacks = 1  # Used as the default when applying this buff if no or negative stacks were specified (min: 1)
    tickrate = 0  # How frequent does this buff tick, in seconds (cannot be lower than 1)

    mods = []  # List of mod objects. See Mod class below for more detail
    cache = {}

    @property
    def ticknum(self):
        """Returns how many ticks this buff has gone through as an integer."""
        x = (time.time() - self.start) / max(1, self.tickrate)
        return int(x)

    @property
    def owner(self):
        """Return this buff's owner (the object its handler is attached to)"""
        if not self.handler:
            return None
        return self.handler.owner

    @property
    def timeleft(self):
        """Returns how much time this buff has left. If -1, it is permanent."""
        _tl = 0
        if not self.start:
            _tl = self.duration
        else:
            _tl = max(-1, self.duration - (time.time() - self.start))
        return _tl

    @property
    def ticking(self) -> bool:
        """Returns if this buff ticks or not (tickrate => 1)"""
        return self.tickrate >= 1

    @property
    def stacking(self) -> bool:
        """Returns if this buff stacks or not (maxstacks > 1)"""
        return self.maxstacks > 1

    def __init__(self, handler, buffkey, cache) -> None:
        """
        Args:
            handler:    The handler this buff is attached to
            buffkey:    The key this buff uses on the cache
            cache:      The cache dictionary (what you get if you use `handler.buffcache.get(key)`)
        """
        required = {"handler": handler, "buffkey": buffkey, "cache": cache}
        self.__dict__.update(cache)
        self.__dict__.update(required)
        # Init hook
        self.at_init()

    def __setattr__(self, attr, value):
        if attr in self.cache:
            if attr == "tickrate":
                value = max(0, value)
            self.handler.buffcache[self.buffkey][attr] = value
        super().__setattr__(attr, value)

    def conditional(self, *args, **kwargs):
        """Hook function for conditional evaluation.

        This must return True for a buff to apply modifiers, trigger effects, or tick."""
        return True

    # region helper methods
    def remove(self, loud=True, expire=False, context=None):
        """Helper method which removes this buff from its handler. Use dispel if you are dispelling it instead.

        Args:
            loud:   (optional) Whether to call at_remove or not (default: True)
            expire: (optional) Whether to call at_expire or not (default: False)
            delay:  (optional) How long you want to delay the remove call for
            context:    (optional) A dictionary you wish to pass to the at_remove/at_expire method as kwargs
        """
        if not context:
            context = {}
        self.handler.remove(self.buffkey, loud=loud, expire=expire, context=context)

    def dispel(self, loud=True, delay=0, context=None):
        """Helper method which dispels this buff (removes and calls at_dispel).

        Args:
            loud:   (optional) Whether to call at_remove or not (default: True)
            delay:  (optional) How long you want to delay the remove call for
            context:    (optional) A dictionary you wish to pass to the at_remove/at_dispel method as kwargs
        """
        if not context:
            context = {}
        self.handler.remove(self.buffkey, loud=loud, dispel=True, delay=delay, context=context)

    def pause(self, context=None):
        """Helper method which pauses this buff on its handler.

        Args:
            context:    (optional) A dictionary you wish to pass to the at_pause method as kwargs"""
        if not context:
            context = {}
        self.handler.pause(self.buffkey, context)

    def unpause(self, context=None):
        """Helper method which unpauses this buff on its handler.

        Args:
            context:    (optional) A dictionary you wish to pass to the at_unpause method as kwargs
        """
        if not context:
            context = {}
        self.handler.unpause(self.buffkey, context)

    def reset(self):
        """Resets the buff start time as though it were just applied; functionally identical to a refresh"""
        self.start = time.time()
        self.handler.buffcache[self.buffkey]["start"] = time.time()

    def update_cache(self, to_cache: dict):
        """Updates this buff's cache using the given values, both internally (this instance) and on the handler.

        Args:
            to_cache:   The dictionary of values you want to add to the cache"""
        if not isinstance(to_cache, dict):
            raise TypeError
        _cache = dict(self.handler.buffcache[self.buffkey])
        _cache.update(to_cache)
        self.cache = _cache
        self.handler.buffcache[self.buffkey] = _cache

    # endregion

    # region hook methods
    def at_init(self, *args, **kwargs):
        """Hook function called when this buff object is initialized."""
        pass

    def at_apply(self, *args, **kwargs):
        """Hook function to run when this buff is applied to an object."""
        pass

    def at_remove(self, *args, **kwargs):
        """Hook function to run when this buff is removed from an object."""
        pass

    def at_dispel(self, *args, **kwargs):
        """Hook function to run when this buff is dispelled from an object (removed by someone other than the buff holder)."""
        pass

    def at_expire(self, *args, **kwargs):
        """Hook function to run when this buff expires from an object."""
        pass

    def at_pre_check(self, *args, **kwargs):
        """Hook function to run before this buff's modifiers are checked."""
        pass

    def at_post_check(self, *args, **kwargs):
        """Hook function to run after this buff's mods are checked."""
        pass

    def at_trigger(self, trigger: str, *args, **kwargs):
        """Hook for the code you want to run whenever the effect is triggered.
        Passes the trigger string to the function, so you can have multiple
        triggers on one buff."""
        pass

    def at_tick(self, initial: bool, *args, **kwargs):
        """Hook for actions that occur per-tick, a designer-set sub-duration.
        `initial` tells you if it's the first tick that happens (when a buff is applied)."""
        pass

    def at_pause(self, *args, **kwargs):
        """Hook for when this buff is paused"""
        pass

    def at_unpause(self, *args, **kwargs):
        """Hook for when this buff is unpaused."""
        pass

    # endregion


class Mod:
    """A single stat mod object. One buff or trait can hold multiple mods, for the same or different stats."""

    stat = "null"  # The stat string that is checked to see if this mod should be applied
    value = 0  # Buff's value
    perstack = 0  # How much additional value is added to the buff per stack
    modifier = "add"  # The modifier the buff applies. 'add' or 'mult'

    def __init__(self, stat: str, modifier: str, value, perstack=0.0) -> None:
        """
        Args:
            stat:       The stat the buff affects. Normally matches the object attribute name
            mod:        The modifier the buff applies. "add" for add/sub or "mult" for mult/div
            value:      The value of the modifier
            perstack:   How much is added to the base, per stack (including first)."""
        self.stat = stat
        self.modifier = modifier
        self.value = value
        self.perstack = perstack


class BuffHandler:
    ownerref = None
    dbkey = "buffs"
    autopause = False
    _owner = None

    def __init__(self, owner, dbkey=dbkey, autopause=autopause):
        """
        Args:
            owner:  The object this handler is attached to
            dbkey:  (optional) The string key of the db attribute to use for the buff cache
            autopause:  (optional) Whether this handler autopauses playtime buffs on owning object's unpuppet
        """
        self.ownerref = owner.dbref
        self.dbkey = dbkey
        self.autopause = autopause
        if autopause:
            self._validate_state()
            signals.SIGNAL_OBJECT_POST_UNPUPPET.connect(self._pause_playtime)
            signals.SIGNAL_OBJECT_POST_PUPPET.connect(self._unpause_playtime)

    # region properties
    @property
    def owner(self):
        """The object this handler is attached to."""
        if self.ownerref:
            _owner = search.search_object(self.ownerref)
        if _owner:
            return _owner[0]
        else:
            return None

    @property
    def buffcache(self):
        """The object attribute we use for the buff cache. Auto-creates if not present."""
        if not self.owner:
            return {}
        if not self.owner.attributes.has(self.dbkey):
            self.owner.attributes.add(self.dbkey, {})
        return self.owner.attributes.get(self.dbkey)

    @property
    def traits(self):
        """All buffs on this handler that modify a stat."""
        _cache = self.all
        _t = {k: buff for k, buff in _cache.items() if buff.mods}
        return _t

    @property
    def effects(self):
        """All buffs on this handler that trigger off an event."""
        _cache = self.all
        _e = {k: buff for k, buff in _cache.items() if buff.triggers}
        return _e

    @property
    def playtime(self):
        """All buffs on this handler that only count down during active playtime."""
        _cache = self.all
        _pt = {k: buff for k, buff in _cache.items() if buff.playtime}
        return _pt

    @property
    def paused(self):
        """All buffs on this handler that are paused."""
        _cache = self.all
        _p = {k: buff for k, buff in _cache.items() if buff.paused}
        return _p

    @property
    def expired(self):
        """All buffs on this handler that have expired (no duration or no stacks)."""
        _cache = self.all
        _e = {
            k: buff
            for k, buff in _cache.items()
            if not buff.paused
            if buff.duration > -1
            if buff.duration < time.time() - buff.start
        }
        _nostacks = {k: buff for k, buff in _cache.items() if buff.stacks <= 0}
        _e.update(_nostacks)
        return _e

    @property
    def visible(self):
        """All buffs on this handler that are visible."""
        _cache = self.all
        _v = {k: buff for k, buff in _cache.items() if buff.visible}
        return _v

    @property
    def all(self):
        """Returns dictionary of instanced buffs equivalent to ALL buffs on this handler,
        regardless of state, type, or anything else."""
        _a = self.get_all()
        return _a

    # endregion

    # region methods
    def add(
        self,
        buff: BaseBuff,
        key: str = None,
        stacks=0,
        duration=None,
        source=None,
        to_cache=None,
        context=None,
        *args,
        **kwargs,
    ):
        """Add a buff to this object, respecting all stacking/refresh/reapplication rules. Takes
        a number of optional parameters to allow for customization.

        Args:
            buff:       The buff class type you wish to add
            key:        (optional) The key you wish to use for this buff; overrides defaults
            stacks:     (optional) The number of stacks you want to add, if the buff is stacking
            duration:   (optional) The amount of time, in seconds, you want the buff to last; overrides defaults
            source:     (optional) The source of this buff. (default: None)
            to_cache:   (optional) A dictionary to store in the buff's cache; does not overwrite default cache keys
            context:    (optional) A dictionary you wish to pass to the at_apply method as kwargs
        """
        if not isinstance(buff, type):
            raise ValueError
        if not context:
            context = {}
        b = {}
        _context = dict(context)

        # Initial cache updating, starting with the class cache attribute and/or to_cache
        if buff.cache:
            b = dict(buff.cache)
        if to_cache:
            b.update(dict(to_cache))

        # Guarantees we stack either at least 1 stack or whatever the class stacks attribute is
        if stacks < 1:
            stacks = min(1, buff.stacks)

        # Create the buff dict that holds a reference and all runtime information.
        b.update(
            {
                "ref": buff,
                "start": time.time(),
                "duration": buff.duration,
                "tickrate": buff.tickrate,
                "prevtick": time.time(),
                "paused": False,
                "stacks": stacks,
                "source": source,
            }
        )

        # Generate the buffkey from the object's dbref and the default buff key.
        # This is the actual key the buff uses on the dictionary
        buffkey = key
        if not buffkey:
            if source:
                mix = str(source.dbref).replace("#", "")
            elif not (buff.unique or buff.refresh) or not source:
                mix = "_ufrf" + str(int((random() * 999999) * 100000))

            buffkey = buff.key if buff.unique is True else buff.key + mix

        # Rules for applying over an existing buff
        if buffkey in self.buffcache.keys():
            existing = dict(self.buffcache[buffkey])
            # Stacking
            if buff.maxstacks > 1:
                b["stacks"] = min(existing["stacks"] + stacks, buff.maxstacks)
            elif buff.maxstacks < 1:
                b["stacks"] = existing["stacks"] + stacks
            # refresh rule for uniques
            if not buff.refresh:
                b["duration"] = existing["duration"]
            # Carrying over old arbitrary cache values
            cur_cache = {k: v for k, v in existing.items() if k not in b.keys()}
            b.update(cur_cache)
        # Setting overloaded duration
        if duration:
            b["duration"] = duration

        # Apply the buff!
        self.buffcache[buffkey] = b

        # Create the buff instance and run the on-application hook method
        instance: BaseBuff = buff(self, buffkey, b)
        instance.at_apply(**_context)
        if instance.ticking:
            tick_buff(self, buffkey, _context)

        # Clean up the buff at the end of its duration through a delayed cleanup call
        if b["duration"] > -1:
            utils.delay(b["duration"], self.cleanup, persistent=True)

    # region removers
    def remove(self, key, stacks=0, loud=True, dispel=False, expire=False, context=None):
        """Remove a buff or effect with matching key from this object. Normally calls at_remove,
        calls at_expire if the buff expired naturally, and optionally calls at_dispel. Can also
        remove stacks instead of the entire buff (still calls at_remove). Typically called via a helper method
        on the buff instance, or other methods on the handler.

        Args:
            key:        The buff key
            loud:       (optional) Calls at_remove when True. (default: True)
            dispel:     (optional) Calls at_dispel when True. (default: False)
            expire:     (optional) Calls at_expire when True. (default: False)
            context:    (optional) A dictionary you wish to pass to the at_remove/at_dispel/at_expire method as kwargs
        """
        if not context:
            context = {}
        if key not in self.buffcache:
            return

        buff: BaseBuff = self.buffcache[key]
        instance: BaseBuff = buff["ref"](self, key, buff)

        if loud:
            if dispel:
                instance.at_dispel(**context)
            elif expire:
                instance.at_expire(**context)
            instance.at_remove(**context)

        del instance
        if not stacks:
            del self.buffcache[key]
        elif stacks:
            self.buffcache[key]["stacks"] -= stacks
            if self.buffcache[key]["stacks"] <= 0:
                del self.buffcache[key]

    def remove_by_type(
        self,
        bufftype: BaseBuff,
        loud=True,
        dispel=False,
        expire=False,
        context=None,
    ):
        """Removes all buffs of a specified type from this object. Functionally similar to remove, but takes a type instead.

        Args:
            bufftype:   The buff class to remove
            loud:       (optional) Calls at_remove when True. (default: True)
            dispel:     (optional) Calls at_dispel when True. (default: False)
            expire:     (optional) Calls at_expire when True. (default: False)
            context:    (optional) A dictionary you wish to pass to the at_remove/at_dispel/at_expire method as kwargs
        """
        _remove = self.get_by_type(bufftype)
        if not _remove:
            return
        self._remove_via_dict(_remove, loud, dispel, expire, context)

    def remove_by_stat(
        self,
        stat,
        loud=True,
        dispel=False,
        expire=False,
        context=None,
    ):
        """Removes all buffs modifying the specified stat from this object.

        Args:
            stat:       The stat string to search for
            loud:       (optional) Calls at_remove when True. (default: True)
            dispel:     (optional) Calls at_dispel when True. (default: False)
            expire:     (optional) Calls at_expire when True. (default: False)
            context:    (optional) A dictionary you wish to pass to the at_remove/at_dispel/at_expire method as kwargs
        """
        _remove = self.get_by_stat(stat)
        if not _remove:
            return
        self._remove_via_dict(_remove, loud, dispel, expire, context)

    def remove_by_trigger(
        self,
        trigger,
        loud=True,
        dispel=False,
        expire=False,
        context=None,
    ):
        """Removes all buffs with the specified trigger from this object.

        Args:
            trigger:    The stat string to search for
            loud:       (optional) Calls at_remove when True. (default: True)
            dispel:     (optional) Calls at_dispel when True. (default: False)
            expire:     (optional) Calls at_expire when True. (default: False)
            context:    (optional) A dictionary you wish to pass to the at_remove/at_dispel/at_expire method as kwargs
        """
        _remove = self.get_by_trigger(trigger)
        if not _remove:
            return
        self._remove_via_dict(_remove, loud, dispel, expire, context)

    def remove_by_source(
        self,
        source,
        loud=True,
        dispel=False,
        expire=False,
        context=None,
    ):
        """Removes all buffs from the specified source from this object.

        Args:
            source:     The source to search for
            loud:       (optional) Calls at_remove when True. (default: True)
            dispel:     (optional) Calls at_dispel when True. (default: False)
            expire:     (optional) Calls at_expire when True. (default: False)
            context:    (optional) A dictionary you wish to pass to the at_remove/at_dispel/at_expire method as kwargs
        """
        _remove = self.get_by_source(source)
        if not _remove:
            return
        self._remove_via_dict(_remove, loud, dispel, expire, context)

    def remove_by_cachevalue(
        self,
        key,
        value=None,
        loud=True,
        dispel=False,
        expire=False,
        context=None,
    ):
        """Removes all buffs with the cachevalue from this object. Functionally similar to remove, but checks the buff's cache values instead.

        Args:
            key:         The key of the cache value to check
            value:      (optional) The value to match to. If None, merely checks to see if the value exists
            loud:       (optional) Calls at_remove when True. (default: True)
            dispel:     (optional) Calls at_dispel when True. (default: False)
            expire:     (optional) Calls at_expire when True. (default: False)
            context:    (optional) A dictionary you wish to pass to the at_remove/at_dispel/at_expire method as kwargs
        """
        _remove = self.get_by_cachevalue(key, value)
        if not _remove:
            return
        self._remove_via_dict(_remove, loud, dispel, expire, context)

    def clear(self, loud=True, dispel=False, expire=False, context=None):
        """Removes all buffs on this handler"""
        cache = self.all
        self._remove_via_dict(cache, loud, dispel, expire, context)

    # endregion
    # region getters
    def get(self, key: str):
        """If the specified key is on this handler, return the instanced buff. Otherwise return None.
        You should delete this when you're done with it, so that garbage collection doesn't have to.

        Args:
            key:    The key for the buff you wish to get"""
        buff = self.buffcache.get(key)
        if buff:
            return buff["ref"](self, key, buff)
        else:
            return None

    def get_all(self):
        """Returns a dictionary of instanced buffs (all of them) on this handler in the format {buffkey: instance}"""
        _cache = dict(self.buffcache)
        if not _cache:
            return {}
        return {k: buff["ref"](self, k, buff) for k, buff in _cache.items()}

    def get_by_type(self, buff: BaseBuff, to_filter=None):
        """Finds all buffs matching the given type.

        Args:
            buff:       The buff class to search for
            to_filter:  (optional) A dictionary you wish to slice. If not provided, uses the whole buffcache.

        Returns a dictionary of instanced buffs of the specified type in the format {buffkey: instance}.
        """
        _cache = self.get_all() if not to_filter else to_filter
        return {k: _buff for k, _buff in _cache.items() if isinstance(_buff, buff)}

    def get_by_stat(self, stat: str, to_filter=None):
        """Finds all buffs which contain a Mod object that modifies the specified stat.

        Args:
            stat:       The string identifier to find relevant mods
            to_filter:  (optional) A dictionary you wish to slice. If not provided, uses the whole buffcache.

        Returns a dictionary of instanced buffs which modify the specified stat in the format {buffkey: instance}.
        """
        _cache = self.traits if not to_filter else to_filter
        buffs = {k: buff for k, buff in _cache.items() for m in buff.mods if m.stat == stat}
        return buffs

    def get_by_trigger(self, trigger: str, to_filter=None):
        """Finds all buffs with the matching string in their triggers.

        Args:
            trigger:    The string identifier to find relevant buffs
            to_filter:  (optional) A dictionary you wish to slice. If not provided, uses the whole buffcache.

        Returns a dictionary of instanced buffs which fire off the designated trigger, in the format {buffkey: instance}.
        """
        _cache = self.effects if not to_filter else to_filter
        buffs = {k: buff for k, buff in _cache.items() if trigger in buff.triggers}
        return buffs

    def get_by_source(self, source, to_filter=None):
        """Find all buffs with the matching source.

        Args:
            source: The source you want to filter buffs by
            to_filter:  (optional) A dictionary you wish to slice. If not provided, uses the whole buffcache.

        Returns a dictionary of instanced buffs which came from the provided source, in the format {buffkey: instance}.
        """
        _cache = self.all if not to_filter else to_filter
        buffs = {k: buff for k, buff in _cache.items() if buff.source == source}
        return buffs

    def get_by_cachevalue(self, key, value=None, to_filter=None):
        """Find all buffs with a matching {key: value} pair in its cache. Allows you to search buffs by arbitrary cache values

        Args:
            key:    The key of the cache value to check
            value:  (optional) The value to match to. If None, merely checks to see if the value exists
            to_filter:  (optional) A dictionary you wish to slice. If not provided, uses the whole buffcache.

        Returns a dictionary of instanced buffs with cache values matching the specified value, in the format {buffkey: instance}.
        """
        _cache = self.all if not to_filter else to_filter
        if not value:
            buffs = {k: buff for k, buff in _cache.items() if buff.cache.get(key)}
        elif value:
            buffs = {k: buff for k, buff in _cache.items() if buff.cache.get(key) == value}
        return buffs

    # endregion

    def has(self, buff=None) -> bool:
        """Checks if the specified buff type or key exists on the handler.

        Args:
            buff:   The buff to search for. This can be a string (the key) or a class reference (the buff type)

        Returns a bool. If no buff and no key is specified, returns False."""
        if not buff:
            return False
        if not (isinstance(buff, type) or isinstance(buff, str)):
            raise TypeError

        if isinstance(buff, str):
            for k in self.buffcache.keys():
                if k == buff:
                    return True
        if isinstance(buff, type):
            for b in self.buffcache.values():
                if b.get("ref") == buff:
                    return True
        return False

    def check(
        self, value: float, stat: str, loud=True, context=None, trigger=False, strongest=False
    ):
        """Finds all buffs and perks related to a stat and applies their effects.

        Args:
            value:  The value you intend to modify
            stat:   The string that designates which stat buffs you want
            loud:   (optional) Call the buff's at_post_check method after checking (default: True)
            context: (optional) A dictionary you wish to pass to the at_pre_check/at_post_check and conditional methods as kwargs
            trigger: (optional) Trigger buffs with the `stat` string as well. (default: False)
            strongest:  (optional) Applies only the strongest mods of the corresponding stat value (default: False)

        Returns the value modified by relevant buffs."""
        # Buff cleanup to make sure all buffs are valid before processing
        self.cleanup()

        # Find all buffs and traits related to the specified stat.
        if not context:
            context = {}
        applied = self.get_by_stat(stat)
        if not applied:
            return value

        # Run pre-check hooks on related buffs
        for buff in applied.values():
            buff.at_pre_check(**context)

        # Sift out buffs that won't be applying their mods (paused, conditional)
        applied = {
            k: buff for k, buff in applied.items() if buff.conditional(**context) if not buff.paused
        }

        # The mod totals
        calc = self._calculate_mods(stat, applied)

        # The calculated final value
        final = self._apply_mods(value, calc, strongest=strongest)

        # Run the "after check" functions on all relevant buffs
        for buff in applied.values():
            buff: BaseBuff
            if loud:
                buff.at_post_check(**context)
            del buff

        # If you want to, also trigger buffs with the same stat string
        if trigger:
            self.trigger(stat, context)

        return final

    def trigger(self, trigger: str, context: dict = None):
        """Calls the at_trigger method on all buffs with the matching trigger.

        Args:
            trigger:    The string identifier to find relevant buffs. Passed to the at_trigger method.
            context:    (optional) A dictionary you wish to pass to the at_trigger method as kwargs
        """
        self.cleanup()
        _effects = self.get_by_trigger(trigger)
        if not _effects:
            return
        if not context:
            context = {}

        _to_trigger = {
            k: buff
            for k, buff in _effects.items()
            if buff.conditional(**context)
            if not buff.paused
            if trigger in buff.triggers
        }

        # Trigger all buffs whose trigger matches the trigger string
        for buff in _to_trigger.values():
            buff: BaseBuff
            buff.at_trigger(trigger, **context)

    def pause(self, key: str, context=None):
        """Pauses the buff. This excludes it from being checked for mods, triggered, or cleaned up. Used to make buffs 'playtime' instead of 'realtime'.

        Args:
            key:    The key for the buff you wish to pause
            context:    (optional) A dictionary you wish to pass to the at_pause method as kwargs
        """
        if key in self.buffcache.keys():
            # Mark the buff as paused
            buff = dict(self.buffcache.get(key))
            if buff["paused"]:
                return
            if not context:
                context = {}
            buff["paused"] = True

            # Math assignments
            current = time.time()  # Current Time
            start = buff["start"]  # Start
            duration = buff["duration"]  # Duration
            prevtick = buff["prevtick"]  # Previous tick timestamp
            tickrate = buff["tickrate"]  # Buff's tick rate
            end = start + duration  # End

            # Setting "tickleft"
            if buff["ref"].ticking:
                buff["tickleft"] = max(1, tickrate - (current - prevtick))

            # Setting the new duration (if applicable)
            if duration > -1:
                newduration = end - current  # New duration
                if newduration > 0:
                    buff["duration"] = newduration
                else:
                    self.remove(key)

            # Apply new cache info, call pause hook
            self.buffcache[key] = buff
            instance: BaseBuff = buff["ref"](self, key, buff)
            instance.at_pause(**context)

    def unpause(self, key: str, context=None):
        """Unpauses a buff. This makes it visible to the various buff systems again.

        Args:
            key:    The key for the buff you wish to pause
            context:    (optional) A dictionary you wish to pass to the at_unpause method as kwargs
        """
        if key in self.buffcache.keys():
            # Mark the buff as unpaused
            buff = dict(self.buffcache.get(key))
            if not buff["paused"]:
                return
            if not context:
                context = {}
            buff["paused"] = False

            # Math assignments
            tickrate = buff["ref"].tickrate
            if buff["ref"].ticking:
                tickleft = buff["tickleft"]
            current = time.time()  # Current Time

            # Start our new timer, adjust prevtick
            buff["start"] = current
            if buff["ref"].ticking:
                buff["prevtick"] = current - (tickrate - tickleft)

            # Apply new cache info, call hook
            self.buffcache[key] = buff
            instance: BaseBuff = buff["ref"](self, key, buff)
            instance.at_unpause(**context)

            # Set up typical delays (cleanup/ticking)
            if instance.duration > -1:
                utils.delay(buff["duration"], cleanup_buffs, self, persistent=True)
            if instance.ticking:
                utils.delay(
                    tickrate, tick_buff, handler=self, buffkey=key, initial=False, persistent=True
                )

    def view(self, to_filter=None) -> dict:
        """Returns a buff flavor text as a dictionary of tuples in the format {key: (name, flavor)}. Common use for this is a buff readout of some kind.

        Args:
            to_filter:  (optional) The dictionary of buffs to iterate over. If none is provided, returns all buffs (default: None)
        """
        if not isinstance(to_filter, dict):
            raise TypeError
        self.cleanup()
        _cache = self.visible if not to_filter else to_filter
        _flavor = {k: (buff.name, buff.flavor) for k, buff in _cache.items()}
        return _flavor

    def view_modifiers(self, stat: str, context=None):
        """Checks all modifiers of the specified stat without actually applying them. Hits the conditional hook for relevant buffs.

        Args:
            stat:   The mod identifier string to search for
            context:    (optional) A dictionary you wish to pass to the conditional hooks as kwargs

        Returns a nested dictionary. The first layer's keys represent the type of modifier ('add' and 'mult'),
        and the second layer's keys represent the type of value ('total' and 'strongest')."""
        # Buff cleanup to make sure all buffs are valid before processing
        self.cleanup()

        # Find all buffs and traits related to the specified stat.
        if not context:
            context = {}
        applied = self.get_by_stat(stat)
        if not applied:
            return None

        # Sift out buffs that won't be applying their mods (paused, conditional)
        applied = {
            k: buff for k, buff in applied.items() if buff.conditional(**context) if not buff.paused
        }

        # Calculate and return our values dictionary
        calc = self._calculate_mods(stat, applied)
        return calc

    def cleanup(self):
        """Removes expired buffs, ensures pause state is respected."""
        self._validate_state()
        cleanup_buffs(self)

    # region private methods
    def _validate_state(self):
        """Validates the state of paused/unpaused playtime buffs."""
        if not self.autopause:
            return
        if self.owner.has_account:
            self._unpause_playtime()
        elif not self.owner.has_account:
            self._pause_playtime()

    def _pause_playtime(self, sender=owner, **kwargs):
        """Pauses all playtime buffs when attached object is unpuppeted."""
        if sender != self.owner:
            return
        buffs = self.playtime
        if not buffs:
            return
        for buff in buffs.values():
            buff.pause()

    def _unpause_playtime(self, sender=owner, **kwargs):
        """Unpauses all playtime buffs when attached object is puppeted."""
        if sender != self.owner:
            return
        buffs = self.playtime
        if not buffs:
            return
        for buff in buffs.values():
            buff.unpause()
        pass

    def _calculate_mods(self, stat: str, buffs: dict):
        """Calculates the total value of applicable mods.

        Args:
            stat:   The string identifier to search mods for
            buffs:  The dictionary of buffs to calculate mods from

        Returns a nested dictionary. The first layer's keys represent the type of modifier ('add' and 'mult'),
        and the second layer's keys represent the type of value ('total' and 'strongest')."""

        # The base return dictionary. If you update how modifiers are calculated, make sure to update this too, or you will get key errors!
        calculated = {
            "add": {"total": 0, "strongest": 0},
            "mult": {"total": 0, "strongest": 0},
            "div": {"total": 0, "strongest": 0},
        }
        if not buffs:
            return calculated

        for buff in buffs.values():
            for mod in buff.mods:
                buff: BaseBuff
                mod: Mod
                if mod.stat == stat:
                    _modval = mod.value + ((buff.stacks) * mod.perstack)
                    calculated[mod.modifier]["total"] += _modval
                    if _modval > calculated[mod.modifier]["strongest"]:
                        calculated[mod.modifier]["strongest"] = _modval
        return calculated

    def _apply_mods(self, value, calc: dict, strongest=False):
        """Applies modifiers to a value.

        Args:
            value:  The value to modify
            calc:   The dictionary of calculated modifier values (see _calculate_mods)
            strongest:  (optional) Applies only the strongest mods of the corresponding stat value (default: False)

        Returns value modified by the relevant mods."""
        final = value
        if strongest:
            final = (
                (value + calc["add"]["strongest"])
                / max(1, 1.0 + calc["div"]["strongest"])
                * max(0, 1.0 + calc["mult"]["strongest"])
            )
        else:
            final = (
                (value + calc["add"]["total"])
                / max(1, 1.0 + calc["div"]["total"])
                * max(0, 1.0 + calc["mult"]["total"])
            )
        return final

    def _remove_via_dict(self, buffs: dict, loud=True, dispel=False, expire=False, context=None):
        """Removes buffs within the provided dictionary from this handler. Used for remove methods besides the basic remove."""
        if not context:
            context = {}
        if not buffs:
            return
        for k, instance in buffs.items():
            instance: BaseBuff
            if loud:
                if dispel:
                    instance.at_dispel(**context)
                elif expire:
                    instance.at_expire(**context)
                instance.at_remove(**context)
            del instance
            del self.buffcache[k]

    # endregion
    # endregion


class BuffableProperty(AttributeProperty):
    """An example of a way you can extend AttributeProperty to create properties that automatically check buffs for you."""

    def at_get(self, value, obj):
        _value = obj.buffs.check(value, self._key)
        return _value


class CmdBuff(Command):
    """
    Buff a target.

    Usage:
      buff <target> <buff>

    Applies the specified buff to the target. All buffs are defined in the bufflist dictionary on this command.
    """

    key = "buff"
    aliases = ["buff"]
    help_category = "builder"

    bufflist = {"foo": BaseBuff}

    def parse(self):
        self.args = self.args.split()

    def func(self):
        caller = self.caller
        target = None
        now = time.time()

        if self.args:
            target = caller.search(self.args[0])
            caller.ndb.target = target
        elif caller.ndb.target:
            target = caller.ndb.target
        else:
            caller.msg("You need to pick a target to buff.")
            return

        if self.args[1] not in self.bufflist.keys():
            caller.msg("You must pick a valid buff.")
            return

        if target:
            target.buffs.add(self.bufflist[self.args[1]], source=caller)
            pass


def cleanup_buffs(handler: BuffHandler):
    """Cleans up all expired buffs from a handler."""
    _remove = handler.expired
    for v in _remove.values():
        v.remove(expire=True)


def tick_buff(handler: BuffHandler, buffkey: str, context=None, initial=True):
    """Ticks a buff. If a buff's tickrate is 1 or larger, this is called when the buff is applied, and then once per tick cycle.

    Args:
        handler:    The handler managing the ticking buff
        buffkey:    The key of the ticking buff
        context:    (optional) A dictionary you wish to pass to the at_tick method as kwargs
        initial:    (optional) Whether this tick_buff call is the first one. Starts True, changes to False for future ticks
    """
    # Cache a reference and find the buff on the object
    if buffkey not in handler.buffcache.keys():
        return
    if not context:
        context = {}

    # Instantiate the buff and tickrate
    buff: BaseBuff = handler.get(buffkey)
    tr = max(1, buff.tickrate)

    # This stops the old ticking process if you refresh/stack the buff
    if (tr > time.time() - buff.prevtick and initial != True) or buff.paused:
        return

    # Only fire the at_tick methods if the conditional is truthy
    if buff.conditional():
        # Always tick this buff on initial
        if initial:
            buff.at_tick(initial, **context)

        # Tick this buff one last time, then remove
        if buff.duration > -1 and buff.duration <= time.time() - buff.start:
            if tr < time.time() - buff.prevtick:
                buff.at_tick(initial, **context)
            buff.remove(expire=True)
            return

        # Tick this buff on-time
        if tr <= time.time() - buff.prevtick:
            buff.at_tick(initial, **context)

    handler.buffcache[buffkey]["prevtick"] = time.time()
    tr = max(1, buff.tickrate)

    # Recur this function at the tickrate interval, if it didn't stop/fail
    utils.delay(
        tr,
        tick_buff,
        handler=handler,
        buffkey=buffkey,
        context=context,
        initial=False,
        persistent=True,
    )



"""
This defines a the Server's generic session object. This object represents
a connection to the outside world but don't know any details about how the
connection actually happens (so it's the same for telnet, web, ssh etc).

It is stored on the Server side (as opposed to protocol-specific sessions which
are stored on the Portal side)
"""
import time

from django.conf import settings
from django.utils import timezone

from evennia.commands.cmdsethandler import CmdSetHandler
from evennia.comms.models import ChannelDB
from evennia.scripts.monitorhandler import MONITOR_HANDLER
from evennia.typeclasses.attributes import (
    AttributeHandler,
    DbHolder,
    InMemoryAttributeBackend,
)
from evennia.utils import logger
from evennia.utils.utils import class_from_module, lazy_property, make_iter

_GA = object.__getattribute__
_SA = object.__setattr__
_ObjectDB = None
_ANSI = None

_BASE_SESSION_CLASS = class_from_module(settings.BASE_SESSION_CLASS)


# -------------------------------------------------------------
# Server Session
# -------------------------------------------------------------


class ServerSession(_BASE_SESSION_CLASS):
    """
    This class represents an account's session and is a template for
    individual protocols to communicate with Evennia.

    Each account gets a session assigned to them whenever they connect
    to the game server. All communication between game and account goes
    through their session.

    """

    # Determines which order command sets begin to be assembled from.
    # Sessions are usually first.
    cmdset_provider_order = 0
    cmdset_provider_error_order = 50
    cmdset_provider_type = "session"

    def __init__(self):
        """
        Initiate to avoid AttributeErrors down the line

        """
        self.puppet = None
        self.account = None
        self.cmdset_storage_string = ""
        self.cmdset = CmdSetHandler(self, True)

    def __cmdset_storage_get(self):
        return [path.strip() for path in self.cmdset_storage_string.split(",")]

    def __cmdset_storage_set(self, value):
        self.cmdset_storage_string = ",".join(str(val).strip() for val in make_iter(value))

    cmdset_storage = property(__cmdset_storage_get, __cmdset_storage_set)

    def get_cmdset_providers(self) -> dict[str, "CmdSetProvider"]:
        """
        Overrideable method which returns a dictionary of every kind of object which
        has a cmdsethandler linked to this ServerSession, and should participate in cmdset
        merging.

        In all normal cases, that's the Session itself, and possibly an account and puppeted
         object.

        Returns:
            dict[str, CmdSetProvider]: The CmdSetProviders linked to this Object.
        """
        out = {"session": self}
        if self.account:
            out["account"] = self.account
        if self.puppet:
            out["object"] = self.puppet
        return out

    @property
    def id(self):
        return self.sessid

    def at_sync(self):
        """
        This is called whenever a session has been resynced with the
        portal.  At this point all relevant attributes have already
        been set and self.account been assigned (if applicable).

        Since this is often called after a server restart we need to
        set up the session as it was.

        """
        global _ObjectDB
        if not _ObjectDB:
            from evennia.objects.models import ObjectDB as _ObjectDB

        super().at_sync()
        if not self.logged_in:
            # assign the unloggedin-command set.
            self.cmdset_storage = settings.CMDSET_UNLOGGEDIN

        self.cmdset.update(init_mode=True)

        if self.puid:
            # reconnect puppet (puid is only set if we are coming
            # back from a server reload). This does all the steps
            # done in the default @ic command but without any
            # hooks, echoes or access checks.
            obj = _ObjectDB.objects.get(id=self.puid)
            obj.sessions.add(self)
            obj.account = self.account
            self.puid = obj.id
            self.puppet = obj
            # obj.scripts.validate()
            obj.locks.cache_lock_bypass(obj)

    def at_login(self, account):
        """
        Hook called by sessionhandler when the session becomes authenticated.

        Args:
            account (Account): The account associated with the session.

        """
        self.account = account
        self.uid = self.account.id
        self.uname = self.account.username
        self.logged_in = True
        self.conn_time = time.time()
        self.puid = None
        self.puppet = None
        self.cmdset_storage = settings.CMDSET_SESSION

        # Update account's last login time.
        self.account.last_login = timezone.now()
        self.account.save()

        # add the session-level cmdset
        self.cmdset = CmdSetHandler(self, True)

    def at_disconnect(self, reason=None):
        """
        Hook called by sessionhandler when disconnecting this session.

        """
        if self.logged_in:
            account = self.account
            if self.puppet:
                account.unpuppet_object(self)
            uaccount = account
            uaccount.last_login = timezone.now()
            uaccount.save()
            # calling account hook
            account.at_disconnect(reason)
            self.logged_in = False
            if not self.sessionhandler.sessions_from_account(account):
                # no more sessions connected to this account
                account.is_connected = False
            # this may be used to e.g. delete account after disconnection etc
            account.at_post_disconnect()
            # remove any webclient settings monitors associated with this
            # session
            MONITOR_HANDLER.remove(account, "_saved_webclient_options", self.sessid)

    def get_account(self):
        """
        Get the account associated with this session

        Returns:
            account (Account or None): The associated Account.

        """
        return self.account if self.logged_in else None

    def get_puppet(self):
        """
        Get the in-game character associated with this session.

        Returns:
            puppet (Object or None): The puppeted object, if any.

        """
        return self.puppet if self.logged_in else None

    get_character = get_puppet

    def get_puppet_or_account(self):
        """
        Get puppet or account.

        Returns:
            controller (Object or Account): The puppet if one exists,
                otherwise return the account.

        """
        if self.logged_in:
            return self.puppet if self.puppet else self.account
        return None

    def log(self, message, channel=True):
        """
        Emits session info to the appropriate outputs and info channels.

        Args:
            message (str): The message to log.
            channel (bool, optional): Log to the CHANNEL_CONNECTINFO channel
                in addition to the server log.

        """
        cchan = channel and settings.CHANNEL_CONNECTINFO
        if cchan:
            try:
                cchan = ChannelDB.objects.get_channel(cchan["key"])
                cchan.msg("[%s]: %s" % (cchan.key, message))
            except Exception:
                logger.log_trace()
        logger.log_info(message)

    def get_client_size(self):
        """
        Return eventual eventual width and height reported by the
        client. Note that this currently only deals with a single
        client window (windowID==0) as in a traditional telnet session.

        """
        flags = self.protocol_flags
        # print("session flags:", flags)
        width = flags.get("SCREENWIDTH", {}).get(0, settings.CLIENT_DEFAULT_WIDTH)
        height = flags.get("SCREENHEIGHT", {}).get(0, settings.CLIENT_DEFAULT_HEIGHT)
        return width, height

    def update_session_counters(self, idle=False):
        """
        Hit this when the user enters a command in order to update
        idle timers and command counters.

        """
        # Idle time used for timeout calcs.
        self.cmd_last = time.time()

        # Store the timestamp of the user's last command.
        if not idle:
            # Increment the user's command counter.
            self.cmd_total += 1
            # Account-visible idle time, not used in idle timeout calcs.
            self.cmd_last_visible = self.cmd_last

    def update_flags(self, **kwargs):
        """
        Update the protocol_flags and sync them with Portal.

        Keyword Args:
            protocol_flag (any): A key and value to set in the
                protocol_flags dictionary.

        Notes:
            Since protocols can vary, no checking is done
            as to the existene of the flag or not. The input
            data should have been validated before this call.

        """
        if kwargs:
            self.protocol_flags.update(kwargs)
            self.sessionhandler.session_portal_sync(self)

    def data_out(self, **kwargs):
        """
        Sending data from Evennia->Client

        Keyword Args:
            text (str or tuple)
            any (str or tuple): Send-commands identified
                by their keys. Or "options", carrying options
                for the protocol(s).

        """
        self.sessionhandler.data_out(self, **kwargs)

    def data_in(self, **kwargs):
        """
        Receiving data from the client, sending it off to
        the respective inputfuncs.

        Keyword Args:
            kwargs (any): Incoming data from protocol on
                the form `{"commandname": ((args), {kwargs}),...}`
        Notes:
            This method is here in order to give the user
            a single place to catch and possibly process all incoming data from
            the client. It should usually always end by sending
            this data off to `self.sessionhandler.call_inputfuncs(self, **kwargs)`.
        """
        self.sessionhandler.call_inputfuncs(self, **kwargs)

    def msg(self, text=None, **kwargs):
        """
        Wrapper to mimic msg() functionality of Objects and Accounts.

        Args:
            text (str): String input.

        Keyword Args:
            any (str or tuple): Send-commands identified
                by their keys. Or "options", carrying options
                for the protocol(s).

        """
        # this can happen if this is triggered e.g. a command.msg
        # that auto-adds the session, we'd get a kwarg collision.
        kwargs.pop("session", None)
        kwargs.pop("from_obj", None)
        if text is not None:
            self.data_out(text=text, **kwargs)
        else:
            self.data_out(**kwargs)

    def execute_cmd(self, raw_string, session=None, **kwargs):
        """
        Do something as this object. This method is normally never
        called directly, instead incoming command instructions are
        sent to the appropriate inputfunc already at the sessionhandler
        level. This method allows Python code to inject commands into
        this stream, and will lead to the text inputfunc be called.

        Args:
            raw_string (string): Raw command input
            session (Session): This is here to make API consistent with
                Account/Object.execute_cmd. If given, data is passed to
                that Session, otherwise use self.
        Keyword Args:
            Other keyword arguments will be added to the found command
            object instace as variables before it executes.  This is
            unused by default Evennia but may be used to set flags and
            change operating paramaters for commands at run-time.

        """
        # inject instruction into input stream
        kwargs["text"] = ((raw_string,), {})
        self.sessionhandler.data_in(session or self, **kwargs)

    def __eq__(self, other):
        """
        Handle session comparisons

        """
        try:
            return self.address == other.address
        except AttributeError:
            return False

    def __hash__(self):
        """
        Python 3 requires that any class which implements __eq__ must also
        implement __hash__ and that the corresponding hashes for equivalent
        instances are themselves equivalent.

        """
        return hash(self.address)

    def __ne__(self, other):
        try:
            return self.address != other.address
        except AttributeError:
            return True

    def __str__(self):
        """
        String representation of the user session class. We use
        this a lot in the server logs.

        """
        symbol = ""
        if self.logged_in and hasattr(self, "account") and self.account:
            symbol = "(#%s)" % self.account.id
        try:
            if hasattr(self.address, "__iter__"):
                address = ":".join([str(part) for part in self.address])
            else:
                address = self.address
        except Exception:
            address = self.address
        return "%s%s@%s" % (self.uname, symbol, address)

    def __repr__(self):
        return "%s" % str(self)

    # Dummy API hooks for use during non-loggedin operation

    def at_cmdset_get(self, **kwargs):
        """
        Called just before cmdsets on this object are requested by the
        command handler. If changes need to be done on the fly to the
        cmdset before passing them on to the cmdhandler, this is the
        place to do it. This is called also if the object currently
        have no cmdsets.

        Keyword Args:
            caller (Object, Account or Session): The object requesting the cmdsets.
            current (CmdSet): The current merged cmdset.
            force_init (bool): If `True`, force a re-build of the cmdset. (seems unused)
            **kwargs: Arbitrary input for overloads.

        """
        pass

    def get_cmdsets(self, caller, current, **kwargs):
        """
        Called by the CommandHandler to get a list of cmdsets to merge.

        Args:
            caller (obj): The object requesting the cmdsets.
            current (cmdset): The current merged cmdset.
            **kwargs: Arbitrary input for overloads.

        Returns:
            tuple: A tuple of (current, cmdsets), which is probably self.cmdset.current and self.cmdset.cmdset_stack
        """
        return self.cmdset.current, list(self.cmdset.cmdset_stack)

    # Mock db/ndb properties for allowing easy storage on the session
    # (note that no databse is involved at all here. session.db.attr =
    # value just saves a normal property in memory, just like ndb).

    @lazy_property
    def nattributes(self):
        return AttributeHandler(self, InMemoryAttributeBackend)

    @lazy_property
    def attributes(self):
        return self.nattributes

    # @property
    def ndb_get(self):
        """
        A non-persistent store (ndb: NonDataBase). Everything stored
        to this is guaranteed to be cleared when a server is shutdown.
        Syntax is same as for the _get_db_holder() method and
        property, e.g. obj.ndb.attr = value etc.

        """
        try:
            return self._ndb_holder
        except AttributeError:
            self._ndb_holder = DbHolder(self, "nattrhandler", manager_name="nattributes")
            return self._ndb_holder

    # @ndb.setter
    def ndb_set(self, value):
        """
        Stop accidentally replacing the db object

        Args:
            value (any): A value to store in the ndb.

        """
        string = "Cannot assign directly to ndb object! "
        string += "Use ndb.attr=value instead."
        raise Exception(string)

    # @ndb.deleter
    def ndb_del(self):
        """
        Stop accidental deletion.

        """
        raise Exception("Cannot delete the ndb object!")

    ndb = property(ndb_get, ndb_set, ndb_del)
    db = property(ndb_get, ndb_set, ndb_del)

    # Mock access method for the session (there is no lock info
    # at this stage, so we just present a uniform API)
    def access(self, *args, **kwargs):
        """
        Dummy method to mimic the logged-in API.

        """
        return True

    def get_display_name(self, *args, **kwargs):
        if self.puppet:
            return self.puppet.get_display_name(*args, **kwargs)
        elif self.account:
            return self.account.get_display_name(*args, **kwargs)
        else:
            return f"{self.protocol_key}({self.address})"



"""

This is the cmdset for Account (OOC) commands.  These are
stored on the Account object and should thus be able to handle getting
an Account object as caller rather than a Character.

Note - in order for session-rerouting (in MULTISESSION_MODE=2) to
function, all commands in this cmdset should use the self.msg()
command method rather than caller.msg().
"""

from evennia.commands.cmdset import CmdSet
from evennia.commands.default import (
    account,
    admin,
    building,
    comms,
    general,
    help,
    system,
)


class AccountCmdSet(CmdSet):
    """
    Implements the account command set.
    """

    key = "DefaultAccount"
    priority = -10

    def at_cmdset_creation(self):
        "Populates the cmdset"

        # Account-specific commands
        self.add(account.CmdOOCLook())
        self.add(account.CmdIC())
        self.add(account.CmdOOC())
        self.add(account.CmdCharCreate())
        self.add(account.CmdCharDelete())
        # self.add(account.CmdSessions())
        self.add(account.CmdWho())
        self.add(account.CmdOption())
        self.add(account.CmdQuit())
        self.add(account.CmdPassword())
        self.add(account.CmdColorTest())
        self.add(account.CmdQuell())
        self.add(account.CmdStyle())

        # nicks
        self.add(general.CmdNick())

        # testing
        self.add(building.CmdExamine())

        # Help command
        self.add(help.CmdHelp())

        # system commands
        self.add(system.CmdReload())
        self.add(system.CmdReset())
        self.add(system.CmdShutdown())
        self.add(system.CmdPy())

        # Admin commands
        self.add(admin.CmdNewPassword())

        # Comm commands
        self.add(comms.CmdChannel())
        self.add(comms.CmdPage())
        self.add(comms.CmdIRC2Chan())
        self.add(comms.CmdIRCStatus())
        self.add(comms.CmdRSS2Chan())
        self.add(comms.CmdGrapevine2Chan())
        self.add(comms.CmdDiscord2Chan())
        # self.add(comms.CmdChannels())
        # self.add(comms.CmdAddCom())
        # self.add(comms.CmdDelCom())
        # self.add(comms.CmdAllCom())
        # self.add(comms.CmdCdestroy())
        # self.add(comms.CmdChannelCreate())
        # self.add(comms.CmdClock())
        # self.add(comms.CmdCBoot())
        # self.add(comms.CmdCWho())
        # self.add(comms.CmdCdesc())



"""
Custom Evennia admin-site, for better customization of the admin-site
as a whole.

This must be located outside of the admin/ folder because it must be imported
before any of the app-data (which in turn must be imported in the `__init__.py`
of that folder for Django to find them).

"""

from django.conf import settings
from django.contrib import admin
from django.contrib.admin import apps


class EvenniaAdminApp(apps.AdminConfig):
    """
    This is imported in INSTALLED_APPS instead of django.contrib.admin.

    """

    default_site = "evennia.web.utils.adminsite.EvenniaAdminSite"


class EvenniaAdminSite(admin.AdminSite):
    """
    The main admin site root (replacing the default from Django). When doing
    admin.register in the admin/ folder, this is what is being registered to.

    """

    site_header = "Evennia web admin"

    def get_app_list(self, request, app_label=None):
        app_list = super().get_app_list(request, app_label=app_label)
        app_mapping = {app["app_label"]: app for app in app_list}
        out = [
            app_mapping.pop(app_label)
            for app_label in settings.DJANGO_ADMIN_APP_ORDER
            if app_label in app_mapping
        ]
        for app in settings.DJANGO_ADMIN_APP_EXCLUDE:
            app_mapping.pop(app, None)
        out += app_mapping.values()
        return out



# Generated by Django 2.2.16 on 2021-05-14 20:32

from django.db import migrations


def migrate_channel_aliases(apps, schema_editor):
    """
    Note - this migration uses the contemporary ChannelDB rather than
    the apps.get_model version. This allows for using all the
    helper functionality, but introduces a dependency on current code. So
    We catch errors and warn, since this is not something that will be needed
    after doing the first migration.

    """
    from evennia.comms.models import ChannelDB

    # ChannelDB = apps.get_model("comms", "ChannelDB")

    for channel in ChannelDB.objects.all():
        try:
            chan_key = channel.db_key.lower()
            channel_aliases = [chan_key] + [alias.lower() for alias in channel.aliases.all()]
            for subscriber in channel.subscriptions.all():
                nicktuples = subscriber.nicks.get(
                    category="channel", return_tuple=True, return_list=True
                )
                all_aliases = channel_aliases + [
                    tup[2] for tup in nicktuples if tup[3].lower() == chan_key
                ]
                for key_or_alias in all_aliases:
                    channel.add_user_channel_alias(subscriber, key_or_alias)
        except Exception as err:
            # we want to continue gracefully here since this is a data-migration from
            # an old to a new version and doesn't involve schema changes
            print("channel-alias data migration 0019_auto_20210514_2032 skipped: {err}")


class Migration(migrations.Migration):

    dependencies = [
        ("comms", "0018_auto_20191025_0831"),
    ]

    operations = [migrations.RunPython(migrate_channel_aliases)]



# -*- coding: utf-8 -*-


from django.db import migrations, models


def convert_defaults(apps, schema_editor):
    AccountDB = apps.get_model("accounts", "AccountDB")
    for account in AccountDB.objects.filter(db_typeclass_path="src.accounts.account.Account"):
        account.db_typeclass_path = "typeclasses.accounts.Account"
        account.save()


class Migration(migrations.Migration):

    dependencies = [("accounts", "0001_initial")]

    operations = [migrations.RunPython(convert_defaults)]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-02-17 20:39


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scripts", "0007_auto_20150403_2339"), ("comms", "0010_auto_20161206_1912")]

    operations = [
        migrations.AddField(
            model_name="msg",
            name="db_receivers_scripts",
            field=models.ManyToManyField(
                blank=True,
                help_text="script_receivers",
                null=True,
                related_name="receiver_script_set",
                to="scripts.ScriptDB",
            ),
        ),
        migrations.AddField(
            model_name="msg",
            name="db_sender_scripts",
            field=models.ManyToManyField(
                blank=True,
                db_index=True,
                null=True,
                related_name="sender_script_set",
                to="scripts.ScriptDB",
                verbose_name="sender(script)",
            ),
        ),
    ]



"""
General Character commands usually available to all characters
"""
import re

from django.conf import settings

import evennia
from evennia.typeclasses.attributes import NickTemplateInvalid
from evennia.utils import utils

COMMAND_DEFAULT_CLASS = utils.class_from_module(settings.COMMAND_DEFAULT_CLASS)

# limit symbol import for API
__all__ = (
    "CmdHome",
    "CmdLook",
    "CmdNick",
    "CmdInventory",
    "CmdSetDesc",
    "CmdGet",
    "CmdDrop",
    "CmdGive",
    "CmdSay",
    "CmdWhisper",
    "CmdPose",
    "CmdAccess",
)


class CmdHome(COMMAND_DEFAULT_CLASS):
    """
    move to your character's home location

    Usage:
      home

    Teleports you to your home location.
    """

    key = "home"
    locks = "cmd:perm(home) or perm(Builder)"
    arg_regex = r"$"

    def func(self):
        """Implement the command"""
        caller = self.caller
        home = caller.home
        if not home:
            caller.msg("You have no home!")
        elif home == caller.location:
            caller.msg("You are already home!")
        else:
            caller.msg("There's no place like home ...")
            caller.move_to(home, move_type="teleport")


class CmdLook(COMMAND_DEFAULT_CLASS):
    """
    look at location or object

    Usage:
      look
      look <obj>
      look *<account>

    Observes your location or objects in your vicinity.
    """

    key = "look"
    aliases = ["l", "ls"]
    locks = "cmd:all()"
    arg_regex = r"\s|$"

    def func(self):
        """
        Handle the looking.
        """
        caller = self.caller
        if not self.args:
            target = caller.location
            if not target:
                caller.msg("You have no location to look at!")
                return
        else:
            target = caller.search(self.args)
            if not target:
                return
        desc = caller.at_look(target)
        # add the type=look to the outputfunc to make it
        # easy to separate this output in client.
        self.msg(text=(desc, {"type": "look"}), options=None)


class CmdNick(COMMAND_DEFAULT_CLASS):
    """
    define a personal alias/nick by defining a string to
    match and replace it with another on the fly

    Usage:
      nick[/switches] <string> [= [replacement_string]]
      nick[/switches] <template> = <replacement_template>
      nick/delete <string> or number
      nicks

    Switches:
      inputline - replace on the inputline (default)
      object    - replace on object-lookup
      account   - replace on account-lookup
      list      - show all defined aliases (also "nicks" works)
      delete    - remove nick by index in /list
      clearall  - clear all nicks

    Examples:
      nick hi = say Hello, I'm Sarah!
      nick/object tom = the tall man
      nick build $1 $2 = create/drop $1;$2
      nick tell $1 $2=page $1=$2
      nick tm?$1=page tallman=$1
      nick tm\=$1=page tallman=$1

    A 'nick' is a personal string replacement. Use $1, $2, ... to catch arguments.
    Put the last $-marker without an ending space to catch all remaining text. You
    can also use unix-glob matching for the left-hand side <string>:

        * - matches everything
        ? - matches 0 or 1 single characters
        [abcd] - matches these chars in any order
        [!abcd] - matches everything not among these chars
        \= - escape literal '=' you want in your <string>

    Note that no objects are actually renamed or changed by this command - your nicks
    are only available to you. If you want to permanently add keywords to an object
    for everyone to use, you need build privileges and the alias command.

    """

    key = "nick"
    switch_options = ("inputline", "object", "account", "list", "delete", "clearall")
    aliases = ["nickname", "nicks"]
    locks = "cmd:all()"

    def parse(self):
        """
        Support escaping of = with \=
        """
        super().parse()
        args = (self.lhs or "") + (" = %s" % self.rhs if self.rhs else "")
        parts = re.split(r"(?<!\\)=", args, 1)
        self.rhs = None
        if len(parts) < 2:
            self.lhs = parts[0].strip()
        else:
            self.lhs, self.rhs = [part.strip() for part in parts]
        self.lhs = self.lhs.replace("\=", "=")

    def func(self):
        """Create the nickname"""

        def _cy(string):
            "add color to the special markers"
            return re.sub(r"(\$[0-9]+|\*|\?|\[.+?\])", r"|Y\1|n", string)

        caller = self.caller
        switches = self.switches
        nicktypes = [switch for switch in switches if switch in ("object", "account", "inputline")]
        specified_nicktype = bool(nicktypes)
        nicktypes = nicktypes if specified_nicktype else ["inputline"]

        nicklist = (
            utils.make_iter(caller.nicks.get(category="inputline", return_obj=True) or [])
            + utils.make_iter(caller.nicks.get(category="object", return_obj=True) or [])
            + utils.make_iter(caller.nicks.get(category="account", return_obj=True) or [])
        )

        if "list" in switches or self.cmdstring in ("nicks",):
            if not nicklist:
                string = "|wNo nicks defined.|n"
            else:
                table = self.styled_table("#", "Type", "Nick match", "Replacement")
                for inum, nickobj in enumerate(nicklist):
                    _, _, nickvalue, replacement = nickobj.value
                    table.add_row(
                        str(inum + 1), nickobj.db_category, _cy(nickvalue), _cy(replacement)
                    )
                string = "|wDefined Nicks:|n\n%s" % table
            caller.msg(string)
            return

        if "clearall" in switches:
            caller.nicks.clear()
            caller.account.nicks.clear()
            caller.msg("Cleared all nicks.")
            return

        if "delete" in switches or "del" in switches:
            if not self.args or not self.lhs:
                caller.msg("usage nick/delete <nick> or <#num> ('nicks' for list)")
                return
            # see if a number was given
            arg = self.args.lstrip("#")
            oldnicks = []
            if arg.isdigit():
                # we are given a index in nicklist
                delindex = int(arg)
                if 0 < delindex <= len(nicklist):
                    oldnicks.append(nicklist[delindex - 1])
                else:
                    caller.msg("Not a valid nick index. See 'nicks' for a list.")
                    return
            else:
                if not specified_nicktype:
                    nicktypes = ("object", "account", "inputline")
                for nicktype in nicktypes:
                    oldnicks.append(caller.nicks.get(arg, category=nicktype, return_obj=True))

            oldnicks = [oldnick for oldnick in oldnicks if oldnick]
            if oldnicks:
                for oldnick in oldnicks:
                    nicktype = oldnick.category
                    nicktypestr = "%s-nick" % nicktype.capitalize()
                    _, _, old_nickstring, old_replstring = oldnick.value
                    caller.nicks.remove(old_nickstring, category=nicktype)
                    caller.msg(
                        f"{nicktypestr} removed: '|w{old_nickstring}|n' -> |w{old_replstring}|n."
                    )
            else:
                caller.msg("No matching nicks to remove.")
            return

        if not self.rhs and self.lhs:
            # check what a nick is set to
            strings = []
            if not specified_nicktype:
                nicktypes = ("object", "account", "inputline")
            for nicktype in nicktypes:
                nicks = [
                    nick
                    for nick in utils.make_iter(
                        caller.nicks.get(category=nicktype, return_obj=True)
                    )
                    if nick
                ]
                for nick in nicks:
                    _, _, nick, repl = nick.value
                    if nick.startswith(self.lhs):
                        strings.append(f"{nicktype.capitalize()}-nick: '{nick}' -> '{repl}'")
            if strings:
                caller.msg("\n".join(strings))
            else:
                caller.msg(f"No nicks found matching '{self.lhs}'")
            return

        if not self.rhs and self.lhs:
            # check what a nick is set to
            strings = []
            if not specified_nicktype:
                nicktypes = ("object", "account", "inputline")
            for nicktype in nicktypes:
                if nicktype == "account":
                    obj = account
                else:
                    obj = caller
                nicks = utils.make_iter(obj.nicks.get(category=nicktype, return_obj=True))
                for nick in nicks:
                    _, _, nick, repl = nick.value
                    if nick.startswith(self.lhs):
                        strings.append(f"{nicktype.capitalize()}-nick: '{nick}' -> '{repl}'")
            if strings:
                caller.msg("\n".join(strings))
            else:
                caller.msg(f"No nicks found matching '{self.lhs}'")
            return

        if not self.rhs and self.lhs:
            # check what a nick is set to
            strings = []
            if not specified_nicktype:
                nicktypes = ("object", "account", "inputline")
            for nicktype in nicktypes:
                if nicktype == "account":
                    obj = account
                else:
                    obj = caller
                nicks = utils.make_iter(obj.nicks.get(category=nicktype, return_obj=True))
                for nick in nicks:
                    _, _, nick, repl = nick.value
                    if nick.startswith(self.lhs):
                        strings.append(f"{nicktype.capitalize()}-nick: '{nick}' -> '{repl}'")
            if strings:
                caller.msg("\n".join(strings))
            else:
                caller.msg(f"No nicks found matching '{self.lhs}'")
            return

        if not self.args or not self.lhs:
            caller.msg("Usage: nick[/switches] nickname = [realname]")
            return

        # setting new nicks

        nickstring = self.lhs
        replstring = self.rhs

        if replstring == nickstring:
            caller.msg("No point in setting nick same as the string to replace...")
            return

        # check so we have a suitable nick type
        errstring = ""
        string = ""
        for nicktype in nicktypes:
            nicktypestr = f"{nicktype.capitalize()}-nick"
            old_nickstring = None
            old_replstring = None

            oldnick = caller.nicks.get(key=nickstring, category=nicktype, return_obj=True)
            if oldnick:
                _, _, old_nickstring, old_replstring = oldnick.value
            if replstring:
                # creating new nick
                errstring = ""
                if oldnick:
                    if replstring == old_replstring:
                        string += f"\nIdentical {nicktypestr.lower()} already set."
                    else:
                        string += (
                            f"\n{nicktypestr} '|w{old_nickstring}|n' updated to map to"
                            f" '|w{replstring}|n'."
                        )
                else:
                    string += f"\n{nicktypestr} '|w{nickstring}|n' mapped to '|w{replstring}|n'."
                try:
                    caller.nicks.add(nickstring, replstring, category=nicktype)
                except NickTemplateInvalid:
                    caller.msg(
                        "You must use the same $-markers both in the nick and in the replacement."
                    )
                    return
            elif old_nickstring and old_replstring:
                # just looking at the nick
                string += f"\n{nicktypestr} '|w{old_nickstring}|n' maps to '|w{old_replstring}|n'."
                errstring = ""
        string = errstring if errstring else string
        caller.msg(_cy(string))


class CmdInventory(COMMAND_DEFAULT_CLASS):
    """
    view inventory

    Usage:
      inventory
      inv

    Shows your inventory.
    """

    key = "inventory"
    aliases = ["inv", "i"]
    locks = "cmd:all()"
    arg_regex = r"$"

    def func(self):
        """check inventory"""
        items = self.caller.contents
        if not items:
            string = "You are not carrying anything."
        else:
            from evennia.utils.ansi import raw as raw_ansi

            table = self.styled_table(border="header")
            for item in items:
                singular, _ = item.get_numbered_name(1, self.caller)
                table.add_row(
                    f"|C{singular}|n",
                    "{}|n".format(utils.crop(raw_ansi(item.db.desc or ""), width=50) or ""),
                )
            string = f"|wYou are carrying:\n{table}"
        self.msg(text=(string, {"type": "inventory"}))


class CmdGet(COMMAND_DEFAULT_CLASS):
    """
    pick up something

    Usage:
      get <obj>

    Picks up an object from your location and puts it in
    your inventory.
    """

    key = "get"
    aliases = "grab"
    locks = "cmd:all();view:perm(Developer);read:perm(Developer)"
    arg_regex = r"\s|$"

    def func(self):
        """implements the command."""

        caller = self.caller

        if not self.args:
            caller.msg("Get what?")
            return
        obj = caller.search(self.args, location=caller.location)
        if not obj:
            return
        if caller == obj:
            caller.msg("You can't get yourself.")
            return
        if not obj.access(caller, "get"):
            if obj.db.get_err_msg:
                caller.msg(obj.db.get_err_msg)
            else:
                caller.msg("You can't get that.")
            return

        # calling at_pre_get hook method
        if not obj.at_pre_get(caller):
            return

        success = obj.move_to(caller, quiet=True, move_type="get")
        if not success:
            caller.msg("This can't be picked up.")
        else:
            singular, _ = obj.get_numbered_name(1, caller)
            caller.location.msg_contents(f"$You() $conj(pick) up {singular}.", from_obj=caller)
            # calling at_get hook method
            obj.at_get(caller)


class CmdDrop(COMMAND_DEFAULT_CLASS):
    """
    drop something

    Usage:
      drop <obj>

    Lets you drop an object from your inventory into the
    location you are currently in.
    """

    key = "drop"
    locks = "cmd:all()"
    arg_regex = r"\s|$"

    def func(self):
        """Implement command"""

        caller = self.caller
        if not self.args:
            caller.msg("Drop what?")
            return

        # Because the DROP command by definition looks for items
        # in inventory, call the search function using location = caller
        obj = caller.search(
            self.args,
            location=caller,
            nofound_string=f"You aren't carrying {self.args}.",
            multimatch_string=f"You carry more than one {self.args}:",
        )
        if not obj:
            return

        # Call the object script's at_pre_drop() method.
        if not obj.at_pre_drop(caller):
            return

        success = obj.move_to(caller.location, quiet=True, move_type="drop")
        if not success:
            caller.msg("This couldn't be dropped.")
        else:
            singular, _ = obj.get_numbered_name(1, caller)
            caller.location.msg_contents(f"$You() $conj(drop) {singular}.", from_obj=caller)
            # Call the object script's at_drop() method.
            obj.at_drop(caller)


class CmdGive(COMMAND_DEFAULT_CLASS):
    """
    give away something to someone

    Usage:
      give <inventory obj> <to||=> <target>

    Gives an item from your inventory to another person,
    placing it in their inventory.
    """

    key = "give"
    rhs_split = ("=", " to ")  # Prefer = delimiter, but allow " to " usage.
    locks = "cmd:all()"
    arg_regex = r"\s|$"

    def func(self):
        """Implement give"""

        caller = self.caller
        if not self.args or not self.rhs:
            caller.msg("Usage: give <inventory object> = <target>")
            return
        to_give = caller.search(
            self.lhs,
            location=caller,
            nofound_string=f"You aren't carrying {self.lhs}.",
            multimatch_string=f"You carry more than one {self.lhs}:",
        )
        target = caller.search(self.rhs)
        if not (to_give and target):
            return

        singular, _ = to_give.get_numbered_name(1, caller)
        if target == caller:
            caller.msg(f"You keep {singular} to yourself.")
            return
        if not to_give.location == caller:
            caller.msg(f"You are not holding {singular}.")
            return

        # calling at_pre_give hook method
        if not to_give.at_pre_give(caller, target):
            return

        # give object
        success = to_give.move_to(target, quiet=True, move_type="give")
        if not success:
            caller.msg(f"You could not give {singular} to {target.key}.")
        else:
            caller.msg(f"You give {singular} to {target.key}.")
            target.msg(f"{caller.key} gives you {singular}.")
            # Call the object script's at_give() method.
            to_give.at_give(caller, target)


class CmdSetDesc(COMMAND_DEFAULT_CLASS):
    """
    describe yourself

    Usage:
      setdesc <description>

    Add a description to yourself. This
    will be visible to people when they
    look at you.
    """

    key = "setdesc"
    locks = "cmd:all()"
    arg_regex = r"\s|$"

    def func(self):
        """add the description"""

        if not self.args:
            self.msg("You must add a description.")
            return

        self.caller.db.desc = self.args.strip()
        self.msg("You set your description.")


class CmdSay(COMMAND_DEFAULT_CLASS):
    """
    speak as your character

    Usage:
      say <message>

    Talk to those in your current location.
    """

    key = "say"
    aliases = ['"', "'"]
    locks = "cmd:all()"

    # don't require a space after `say/'/"`
    arg_regex = None

    def func(self):
        """Run the say command"""

        caller = self.caller

        if not self.args:
            caller.msg("Say what?")
            return

        speech = self.args

        # Calling the at_pre_say hook on the character
        speech = caller.at_pre_say(speech)

        # If speech is empty, stop here
        if not speech:
            return

        # Call the at_post_say hook on the character
        caller.at_say(speech, msg_self=True)


class CmdWhisper(COMMAND_DEFAULT_CLASS):
    """
    Speak privately as your character to another

    Usage:
      whisper <character> = <message>
      whisper <char1>, <char2> = <message>

    Talk privately to one or more characters in your current location, without
    others in the room being informed.
    """

    key = "whisper"
    locks = "cmd:all()"

    def func(self):
        """Run the whisper command"""

        caller = self.caller

        if not self.lhs or not self.rhs:
            caller.msg("Usage: whisper <character> = <message>")
            return

        receivers = [recv.strip() for recv in self.lhs.split(",")]

        receivers = [caller.search(receiver) for receiver in set(receivers)]
        receivers = [recv for recv in receivers if recv]

        speech = self.rhs
        # If the speech is empty, abort the command
        if not speech or not receivers:
            return

        # Call a hook to change the speech before whispering
        speech = caller.at_pre_say(speech, whisper=True, receivers=receivers)

        # no need for self-message if we are whispering to ourselves (for some reason)
        msg_self = None if caller in receivers else True
        caller.at_say(speech, msg_self=msg_self, receivers=receivers, whisper=True)


class CmdPose(COMMAND_DEFAULT_CLASS):
    """
    strike a pose

    Usage:
      pose <pose text>
      pose's <pose text>

    Example:
      pose is standing by the wall, smiling.
       -> others will see:
      Tom is standing by the wall, smiling.

    Describe an action being taken. The pose text will
    automatically begin with your name.
    """

    key = "pose"
    aliases = [":", "emote"]
    locks = "cmd:all()"
    arg_regex = ""

    # we want to be able to pose without whitespace between
    # the command/alias and the pose (e.g. :pose)
    arg_regex = None

    def parse(self):
        """
        Custom parse the cases where the emote
        starts with some special letter, such
        as 's, at which we don't want to separate
        the caller's name and the emote with a
        space.
        """
        args = self.args
        if args and not args[0] in ["'", ",", ":"]:
            args = " %s" % args.strip()
        self.args = args

    def func(self):
        """Hook function"""
        if not self.args:
            msg = "What do you want to do?"
            self.msg(msg)
        else:
            msg = f"{self.caller.name}{self.args}"
            self.caller.location.msg_contents(text=(msg, {"type": "pose"}), from_obj=self.caller)


class CmdAccess(COMMAND_DEFAULT_CLASS):
    """
    show your current game access

    Usage:
      access

    This command shows you the permission hierarchy and
    which permission groups you are a member of.
    """

    key = "access"
    aliases = ["groups", "hierarchy"]
    locks = "cmd:all()"
    arg_regex = r"$"

    def func(self):
        """Load the permission groups"""

        caller = self.caller
        hierarchy_full = settings.PERMISSION_HIERARCHY
        string = "\n|wPermission Hierarchy|n (climbing):\n %s" % ", ".join(hierarchy_full)

        if self.caller.account.is_superuser:
            cperms = "<Superuser>"
            pperms = "<Superuser>"
        else:
            cperms = ", ".join(caller.permissions.all())
            pperms = ", ".join(caller.account.permissions.all())

        string += "\n|wYour access|n:"
        string += f"\nCharacter |c{caller.key}|n: {cperms}"
        if utils.inherits_from(caller, evennia.DefaultObject):
            string += f"\nAccount |c{caller.account.key}|n: {pperms}"
        caller.msg(string)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-17 20:17


from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("comms", "0011_auto_20170606_1731"), ("comms", "0011_auto_20170217_2039")]

    operations = []



"""
The managers for the custom Account object and permissions.
"""

import datetime

from django.conf import settings
from django.contrib.auth.models import UserManager
from django.utils import timezone

from evennia.server import signals
from evennia.typeclasses.managers import TypeclassManager, TypedObjectManager
from evennia.utils.utils import class_from_module, dbid_to_obj, make_iter

__all__ = ("AccountManager", "AccountDBManager")


#
# Account Manager
#


class AccountDBManager(TypedObjectManager, UserManager):
    """
    This AccountManager implements methods for searching
    and manipulating Accounts directly from the database.

    Evennia-specific search methods (will return Characters if
    possible or a Typeclass/list of Typeclassed objects, whereas
    Django-general methods will return Querysets or database objects):

    dbref (converter)
    dbref_search
    get_dbref_range
    object_totals
    typeclass_search
    num_total_accounts
    get_connected_accounts
    get_recently_created_accounts
    get_recently_connected_accounts
    get_account_from_email
    get_account_from_uid
    get_account_from_name
    account_search (equivalent to evennia.search_account)

    """

    def num_total_accounts(self):
        """
        Get total number of accounts.

        Returns:
            count (int): The total number of registered accounts.

        """
        return self.count()

    def get_connected_accounts(self):
        """
        Get all currently connected accounts.

        Returns:
            count (list): Account objects with currently
                connected sessions.

        """
        return self.filter(db_is_connected=True)

    def get_recently_created_accounts(self, days=7):
        """
        Get accounts recently created.

        Args:
            days (int, optional): How many days in the past "recently" means.

        Returns:
            accounts (list): The Accounts created the last `days` interval.

        """
        end_date = timezone.now()
        tdelta = datetime.timedelta(days)
        start_date = end_date - tdelta
        return self.filter(date_joined__range=(start_date, end_date))

    def get_recently_connected_accounts(self, days=7):
        """
        Get accounts recently connected to the game.

        Args:
            days (int, optional): Number of days backwards to check

        Returns:
            accounts (list): The Accounts connected to the game in the
                last `days` interval.

        """
        end_date = timezone.now()
        tdelta = datetime.timedelta(days)
        start_date = end_date - tdelta
        return self.filter(last_login__range=(start_date, end_date)).order_by("-last_login")

    def get_account_from_email(self, uemail):
        """
        Search account by
        Returns an account object based on email address.

        Args:
            uemail (str): An email address to search for.

        Returns:
            account (Account): A found account, if found.

        """
        return self.filter(email__iexact=uemail)

    def get_account_from_uid(self, uid):
        """
        Get an account by id.

        Args:
            uid (int): Account database id.

        Returns:
            account (Account): The result.

        """
        try:
            return self.get(id=uid)
        except self.model.DoesNotExist:
            return None

    def get_account_from_name(self, uname):
        """
        Get account object based on name.

        Args:
            uname (str): The Account name to search for.

        Returns:
            account (Account): The found account.

        """
        try:
            return self.get(username__iexact=uname)
        except self.model.DoesNotExist:
            return None

    def search_account(self, ostring, exact=True, typeclass=None):
        """
        Searches for a particular account by name or
        database id.

        Args:
            ostring (str or int): A key string or database id.
            exact (bool, optional): Only valid for string matches. If
                `True`, requires exact (non-case-sensitive) match,
                otherwise also match also keys containing the `ostring`
                (non-case-sensitive fuzzy match).
            typeclass (str or Typeclass, optional): Limit the search only to
                accounts of this typeclass.
        Returns:
            Queryset: A queryset (an iterable) with 0, 1 or more matches.

        """
        dbref = self.dbref(ostring)
        if dbref or dbref == 0:
            # dbref search is always exact
            dbref_match = self.search_dbref(dbref)
            if dbref_match:
                return dbref_match

        query = {"username__iexact" if exact else "username__icontains": ostring}
        if typeclass:
            # we accept both strings and actual typeclasses
            if callable(typeclass):
                typeclass = f"{typeclass.__module__}.{typeclass.__name__}"
            else:
                typeclass = str(typeclass)
            query["db_typeclass_path"] = typeclass
        if exact:
            matches = self.filter(**query)
        else:
            matches = self.filter(**query)
        if not matches:
            # try alias match
            matches = self.filter(
                db_tags__db_tagtype__iexact="alias",
                **{"db_tags__db_key__iexact" if exact else "db_tags__db_key__icontains": ostring},
            )
        return matches

    def create_account(
        self,
        key,
        email,
        password,
        typeclass=None,
        is_superuser=False,
        locks=None,
        permissions=None,
        tags=None,
        attributes=None,
        report_to=None,
    ):
        """
        This creates a new account.

        Args:
            key (str): The account's name. This should be unique.
            email (str or None): Email on valid addr@addr.domain form. If
                the empty string, will be set to None.
            password (str): Password in cleartext.

        Keyword Args:
            typeclass (str): The typeclass to use for the account.
            is_superuser (bool): Whether or not this account is to be a superuser
            locks (str): Lockstring.
            permission (list): List of permission strings.
            tags (list): List of Tags on form `(key, category[, data])`
            attributes (list): List of Attributes on form
                 `(key, value [, category, [,lockstring [, default_pass]]])`
            report_to (Object): An object with a msg() method to report
                errors to. If not given, errors will be logged.

        Returns:
            Account: The newly created Account.
        Raises:
            ValueError: If `key` already exists in database.


        Notes:
            Usually only the server admin should need to be superuser, all
            other access levels can be handled with more fine-grained
            permissions or groups. A superuser bypasses all lock checking
            operations and is thus not suitable for play-testing the game.

        """
        typeclass = typeclass if typeclass else settings.BASE_ACCOUNT_TYPECLASS
        locks = make_iter(locks) if locks is not None else None
        permissions = make_iter(permissions) if permissions is not None else None
        tags = make_iter(tags) if tags is not None else None
        attributes = make_iter(attributes) if attributes is not None else None

        if isinstance(typeclass, str):
            # a path is given. Load the actual typeclass.
            typeclass = class_from_module(typeclass, settings.TYPECLASS_PATHS)

        # setup input for the create command. We use AccountDB as baseclass
        # here to give us maximum freedom (the typeclasses will load
        # correctly when each object is recovered).

        if not email:
            email = None
        if self.model.objects.filter(username__iexact=key):
            raise ValueError("An Account with the name '%s' already exists." % key)

        # this handles a given dbref-relocate to an account.
        report_to = dbid_to_obj(report_to, self.model)

        # create the correct account entity, using the setup from
        # base django auth.
        now = timezone.now()
        email = typeclass.objects.normalize_email(email)
        new_account = typeclass(
            username=key,
            email=email,
            is_staff=is_superuser,
            is_superuser=is_superuser,
            last_login=now,
            date_joined=now,
        )
        if password is not None:
            # the password may be None for 'fake' accounts, like bots
            valid, error = new_account.validate_password(password, new_account)
            if not valid:
                raise error

            new_account.set_password(password)

        new_account._createdict = dict(
            locks=locks,
            permissions=permissions,
            report_to=report_to,
            tags=tags,
            attributes=attributes,
        )
        # saving will trigger the signal that calls the
        # at_first_save hook on the typeclass, where the _createdict
        # can be used.
        new_account.save()

        # note that we don't send a signal here, that is sent from the Account.create helper method
        # instead.

        return new_account

    # back-compatibility alias
    account_search = search_account


class AccountManager(AccountDBManager, TypeclassManager):
    pass



# -*- coding: utf-8 -*-
"""
This sub-package contains Evennia's command system. It handles
everything related to parsing input from the account, building cmdsets
and executing the code associated with a found command class.

commands.default contains all the default "mux-like" commands of
Evennia.

"""



"""
A *lock* defines access to a particular subsystem or property of
Evennia. For example, the "owner" property can be impmemented as a
lock. Or the disability to lift an object or to ban users.


A lock consists of three parts:

 - access_type - this defines what kind of access this lock regulates. This
   just a string.
 - function call - this is one or many calls to functions that will determine
   if the lock is passed or not.
 - lock function(s). These are regular python functions with a special
   set of allowed arguments. They should always return a boolean depending
   on if they allow access or not.

A lock function is defined by existing in one of the modules
listed by settings.LOCK_FUNC_MODULES. It should also always
take four arguments looking like this:

   funcname(accessing_obj, accessed_obj, *args, **kwargs):
        [...]

The accessing object is the object wanting to gain access.
The accessed object is the object this lock resides on
args and kwargs will hold optional arguments and/or keyword arguments
to the function as a list and a dictionary respectively.

Example:

   perm(accessing_obj, accessed_obj, *args, **kwargs):
       "Checking if the object has a particular, desired permission"
       if args:
           desired_perm = args[0]
           return desired_perm in accessing_obj.permissions.all()
       return False

Lock functions should most often be pretty general and ideally possible to
re-use and combine in various ways to build clever locks.



Lock definition ("Lock string")

A lock definition is a string with a special syntax. It is added to
each object's lockhandler, making that lock available from then on.

The lock definition looks like this:

 'access_type:[NOT] func1(args)[ AND|OR][NOT] func2() ...'

That is, the access_type, a colon followed by calls to lock functions
combined with AND or OR. NOT negates the result of the following call.

Example:

 We want to limit who may edit a particular object (let's call this access_type
for 'edit', it depends on what the command is looking for). We want this to
only work for those with the Permission 'Builder'. So we use our lock
function above and define it like this:

  'edit:perm(Builder)'

Here, the lock-function perm() will be called with the string
'Builder' (accessing_obj and accessed_obj are added automatically,
you only need to add the args/kwargs, if any).

If we wanted to make sure the accessing object was BOTH a Builder and a
GoodGuy, we could use AND:

  'edit:perm(Builder) AND perm(GoodGuy)'

To allow EITHER Builder and GoodGuys, we replace AND with OR. perm() is just
one example, the lock function can do anything and compare any properties of
the calling object to decide if the lock is passed or not.

  'lift:attrib(very_strong) AND NOT attrib(bad_back)'

To make these work, add the string to the lockhandler of the object you want
to apply the lock to:

  obj.lockhandler.add('edit:perm(Builder)')

From then on, a command that wants to check for 'edit' access on this
object would do something like this:

  if not target_obj.lockhandler.has_perm(caller, 'edit'):
      caller.msg("Sorry, you cannot edit that.")

All objects also has a shortcut called 'access' that is recommended to
use instead:

  if not target_obj.access(caller, 'edit'):
      caller.msg("Sorry, you cannot edit that.")


Permissions

Permissions are just text strings stored in a comma-separated list on
typeclassed objects. The default perm() lock function uses them,
taking into account settings.PERMISSION_HIERARCHY. Also, the
restricted @perm command sets them, but otherwise they are identical
to any other identifier you can use.

"""

import re

from django.conf import settings
from django.utils.translation import gettext as _

import evennia
from evennia.utils import logger, utils

__all__ = ("LockHandler", "LockException")

WARNING_LOG = settings.LOCKWARNING_LOG_FILE
_LOCK_HANDLER = None


#
# Exception class. This will be raised
# by errors in lock definitions.
#


class LockException(Exception):
    """
    Raised during an error in a lock.

    """

    pass


#
# Cached lock functions
#

_LOCKFUNCS = {}


def _cache_lockfuncs():
    """
    Updates the cache.

    """
    global _LOCKFUNCS
    _LOCKFUNCS = {}
    for modulepath in settings.LOCK_FUNC_MODULES:
        _LOCKFUNCS.update(utils.callables_from_module(modulepath))


#
# pre-compiled regular expressions
#


_RE_FUNCS = re.compile(r"\w+\([^)]*\)")
_RE_SEPS = re.compile(r"(?<=[ )])AND(?=\s)|(?<=[ )])OR(?=\s)|(?<=[ )])NOT(?=\s)")
_RE_OK = re.compile(r"%s|and|or|not")


#
#
# Lock handler
#
#


class LockHandler:
    """
    This handler should be attached to all objects implementing
    permission checks, under the property 'lockhandler'.

    """

    def __init__(self, obj):
        """
        Loads and pre-caches all relevant locks and their functions.

        Args:
            obj (object): The object on which the lockhandler is
            defined.

        """
        if not _LOCKFUNCS:
            _cache_lockfuncs()
        self.obj = obj
        self.locks = {}
        try:
            self.reset()
        except LockException as err:
            logger.log_trace(err)

    def __str__(self):
        return ";".join(self.locks[key][2] for key in sorted(self.locks))

    def _log_error(self, message):
        "Try to log errors back to object"
        raise LockException(message)

    def _parse_lockstring(self, storage_lockstring):
        """
        Helper function. This is normally only called when the
        lockstring is cached and does preliminary checking.  locks are
        stored as a string

            atype:[NOT] lock()[[ AND|OR [NOT] lock()[...]];atype...

        Args:
            storage_locksring (str): The lockstring to parse.

        """
        locks = {}
        if not storage_lockstring:
            return locks
        duplicates = 0
        elist = []  # errors
        wlist = []  # warnings
        for raw_lockstring in storage_lockstring.split(";"):
            if not raw_lockstring:
                continue
            lock_funcs = []
            try:
                access_type, rhs = (part.strip() for part in raw_lockstring.split(":", 1))
            except ValueError:
                logger.log_trace()
                return locks

            # parse the lock functions and separators
            funclist = _RE_FUNCS.findall(rhs)
            evalstring = rhs
            for pattern in ("AND", "OR", "NOT"):
                evalstring = re.sub(r"\b%s\b" % pattern, pattern.lower(), evalstring)
            nfuncs = len(funclist)
            for funcstring in funclist:
                funcname, rest = (part.strip().strip(")") for part in funcstring.split("(", 1))
                func = _LOCKFUNCS.get(funcname, None)
                if not callable(func):
                    elist.append(
                        _("Lock: lock-function '{lockfunc}' is not available.").format(
                            lockfunc=funcstring
                        )
                    )
                    continue
                args = list(arg.strip() for arg in rest.split(",") if arg and "=" not in arg)
                kwargs = dict(
                    [
                        (part.strip() for part in arg.split("=", 1))
                        for arg in rest.split(",")
                        if arg and "=" in arg
                    ]
                )
                lock_funcs.append((func, args, kwargs))
                evalstring = evalstring.replace(funcstring, "%s")
            if len(lock_funcs) < nfuncs:
                continue
            try:
                # purge the eval string of any superfluous items, then test it
                evalstring = " ".join(_RE_OK.findall(evalstring))
                eval(evalstring % tuple(True for func in funclist), {}, {})
            except Exception:
                elist.append(
                    _("Lock: definition '{lock_string}' has syntax errors.").format(
                        lock_string=raw_lockstring
                    )
                )
                continue
            if access_type in locks:
                duplicates += 1
                wlist.append(
                    _(
                        "LockHandler on {obj}: access type '{access_type}' "
                        "changed from '{source}' to '{goal}' ".format(
                            obj=self.obj,
                            access_type=access_type,
                            source=locks[access_type][2],
                            goal=raw_lockstring,
                        )
                    )
                )
            locks[access_type] = (evalstring, tuple(lock_funcs), raw_lockstring)
        if wlist and WARNING_LOG:
            # a warning text was set, it's not an error, so only report
            logger.log_file("\n".join(wlist), WARNING_LOG)
        if elist:
            # an error text was set, raise exception.
            raise LockException("\n".join(elist))
        # return the gathered locks in an easily executable form
        return locks

    def _cache_locks(self, storage_lockstring):
        """
        Store data

        """
        self.locks = self._parse_lockstring(storage_lockstring)

    def _save_locks(self):
        """
        Store locks to obj

        """
        self.obj.lock_storage = ";".join([tup[2] for tup in self.locks.values()])

    def cache_lock_bypass(self, obj):
        """
        We cache superuser bypass checks here for efficiency. This
        needs to be re-run when an account is assigned to a character.
        We need to grant access to superusers. We need to check both
        directly on the object (accounts), through obj.account and using
        the get_account() method (this sits on serversessions, in some
        rare cases where a check is done before the login process has
        yet been fully finalized)

        Args:
            obj (object): This is checked for the `is_superuser` property.

        """
        self.lock_bypass = hasattr(obj, "is_superuser") and obj.is_superuser

    def add(self, lockstring, validate_only=False):
        """
        Add a new lockstring to handler.

        Args:
            lockstring (str or list): A string on the form
                `"<access_type>:<functions>"`.  Multiple access types
                should be separated by semicolon (`;`). Alternatively,
                a list with lockstrings.
            validate_only (bool, optional): If True, validate the lockstring but
                don't actually store it.
        Returns:
            success (bool): The outcome of the addition, `False` on
                error. If `validate_only` is True, this will be a tuple
                (bool, error), for pass/fail and a string error.

        """
        if isinstance(lockstring, str):
            lockdefs = [
                stripped for lockdef in lockstring.split(";") if (stripped := lockdef.strip())
            ]
        else:
            lockdefs = [
                stripped
                for locks in lockstring
                for lockdef in locks.split(";")
                if (stripped := lockdef.strip())
            ]
            lockstring = ";".join(lockdefs)

        err = ""
        # sanity checks
        for lockdef in lockdefs:
            if ":" not in lockdef:
                err = _("Lock: '{lockdef}' contains no colon (:).").format(lockdef=lockdef)
                if validate_only:
                    return False, err
                else:
                    self._log_error(err)
                    return False
            access_type, rhs = [part.strip() for part in lockdef.split(":", 1)]
            if not access_type:
                err = _(
                    "Lock: '{lockdef}' has no access_type " "(left-side of colon is empty)."
                ).format(lockdef=lockdef)
                if validate_only:
                    return False, err
                else:
                    self._log_error(err)
                    return False
            if rhs.count("(") != rhs.count(")"):
                err = _("Lock: '{lockdef}' has mismatched parentheses.").format(lockdef=lockdef)
                if validate_only:
                    return False, err
                else:
                    self._log_error(err)
                    return False
            if not _RE_FUNCS.findall(rhs):
                err = _("Lock: '{lockdef}' has no valid lock functions.").format(lockdef=lockdef)
                if validate_only:
                    return False, err
                else:
                    self._log_error(err)
                    return False
        if validate_only:
            return True, None
        # get the lock string
        storage_lockstring = self.obj.lock_storage
        if storage_lockstring:
            storage_lockstring = storage_lockstring + ";" + lockstring
        else:
            storage_lockstring = lockstring
        # cache the locks will get rid of eventual doublets
        self._cache_locks(storage_lockstring)
        self._save_locks()
        return True

    def validate(self, lockstring):
        """
        Validate lockstring syntactically, without saving it.

        Args:
            lockstring (str): Lockstring to validate.
        Returns:
            valid (bool): If validation passed or not.

        """
        return self.add(lockstring, validate_only=True)

    def replace(self, lockstring):
        """
        Replaces the lockstring entirely.

        Args:
            lockstring (str): The new lock definition.

        Return:
            success (bool): False if an error occurred.

        Raises:
            LockException: If a critical error occurred.
                If so, the old string is recovered.

        """
        old_lockstring = str(self)
        self.clear()
        try:
            return self.add(lockstring)
        except LockException:
            self.add(old_lockstring)
            raise

    def get(self, access_type=None):
        """
        Get the full lockstring or the lockstring of a particular
        access type.

        Args:
            access_type (str, optional):

        Returns:
            lockstring (str): The matched lockstring, or the full
                lockstring if no access_type was given.
        """

        if access_type:
            return self.locks.get(access_type, ["", "", ""])[2]
        return str(self)

    def all(self):
        """
        Return all lockstrings

        Returns:
            lockstrings (list): All separate lockstrings

        """
        return str(self).split(";")

    def remove(self, access_type):
        """
        Remove a particular lock from the handler

        Args:
            access_type (str): The type of lock to remove.

        Returns:
            success (bool): If the access_type was not found
                in the lock, this returns `False`.

        """
        if access_type in self.locks:
            del self.locks[access_type]
            self._save_locks()
            return True
        return False

    delete = remove  # alias for historical reasons

    def clear(self):
        """
        Remove all locks in the handler.

        """
        self.locks = {}
        self.lock_storage = ""
        self._save_locks()

    def reset(self):
        """
        Set the reset flag, so the the lock will be re-cached at next
        checking.  This is usually called by @reload.

        """
        self._cache_locks(self.obj.lock_storage)
        self.cache_lock_bypass(self.obj)

    def append(self, access_type, lockstring, op="or"):
        """
        Append a lock definition to access_type if it doesn't already exist.

        Args:
            access_type (str): Access type.
            lockstring (str): A valid lockstring, without the operator to
                link it to an eventual existing lockstring.
            op (str): An operator 'and', 'or', 'and not', 'or not' used
                for appending the lockstring to an existing access-type.
        Note:
            The most common use of this method is for use in commands where
            the user can specify their own lockstrings. This method allows
            the system to auto-add things like Admin-override access.

        """
        old_lockstring = self.get(access_type)
        if not lockstring.strip().lower() in old_lockstring.lower():
            lockstring = "{old} {op} {new}".format(
                old=old_lockstring, op=op, new=lockstring.strip()
            )
            self.add(lockstring)

    def check(self, accessing_obj, access_type, default=False, no_superuser_bypass=False):
        """
        Checks a lock of the correct type by passing execution off to
        the lock function(s).

        Args:
            accessing_obj (object): The object seeking access.
            access_type (str): The type of access wanted.
            default (bool, optional): If no suitable lock type is
                found, default to this result.
            no_superuser_bypass (bool): Don't use this unless you
                really, really need to, it makes supersusers susceptible
                to the lock check.

        Notes:
            A lock is executed in the follwoing way:

            Parsing the lockstring, we (during cache) extract the valid
            lock functions and store their function objects in the right
            order along with their args/kwargs. These are now executed in
            sequence, creating a list of True/False values. This is put
            into the evalstring, which is a string of AND/OR/NOT entries
            separated by placeholders where each function result should
            go. We just put those results in and evaluate the string to
            get a final, combined True/False value for the lockstring.

            The important bit with this solution is that the full
            lockstring is never blindly evaluated, and thus there (should
            be) no way to sneak in malign code in it. Only "safe" lock
            functions (as defined by your settings) are executed.

        """
        try:
            # check if the lock should be bypassed (e.g. superuser status)
            if accessing_obj.locks.lock_bypass and not no_superuser_bypass:
                return True
        except AttributeError:
            # happens before session is initiated.
            if not no_superuser_bypass and (
                (hasattr(accessing_obj, "is_superuser") and accessing_obj.is_superuser)
                or (
                    utils.inherits_from(accessing_obj, evennia.DefaultObject)
                    and hasattr(accessing_obj.account, "is_superuser")
                    and accessing_obj.account.is_superuser
                )
                or (
                    hasattr(accessing_obj, "get_account")
                    and (
                        not accessing_obj.get_account() or accessing_obj.get_account().is_superuser
                    )
                )
            ):
                return True

        # no superuser or bypass -> normal lock operation
        if access_type in self.locks:
            # we have a lock, test it.
            evalstring, func_tup, raw_string = self.locks[access_type]
            # execute all lock funcs in the correct order, producing a tuple of True/False results.
            true_false = tuple(
                bool(tup[0](accessing_obj, self.obj, *tup[1], access_type=access_type, **tup[2]))
                for tup in func_tup
            )
            # the True/False tuple goes into evalstring, which combines them
            # with AND/OR/NOT in order to get the final result.
            return eval(evalstring % true_false)
        else:
            return default

    def _eval_access_type(self, accessing_obj, locks, access_type):
        """
        Helper method for evaluating the access type using eval().

        Args:
            accessing_obj (object): Object seeking access.
            locks (dict): The pre-parsed representation of all access-types.
            access_type (str): An access-type key to evaluate.

        """
        evalstring, func_tup, raw_string = locks[access_type]
        true_false = tuple(tup[0](accessing_obj, self.obj, *tup[1], **tup[2]) for tup in func_tup)
        return eval(evalstring % true_false)

    def check_lockstring(
        self, accessing_obj, lockstring, no_superuser_bypass=False, default=False, access_type=None
    ):
        """
        Do a direct check against a lockstring ('atype:func()..'),
        without any intermediary storage on the accessed object.

        Args:
            accessing_obj (object or None): The object seeking access.
                Importantly, this can be left unset if the lock functions
                don't access it, no updating or storage of locks are made
                against this object in this method.
            lockstring (str): Lock string to check, on the form
                `"access_type:lock_definition"` where the `access_type`
                part can potentially be set to a dummy value to just check
                a lock condition.
            no_superuser_bypass  (bool, optional): Force superusers to heed lock.
            default (bool, optional): Fallback result to use if `access_type` is set
                but no such `access_type` is found in the given `lockstring`.
            access_type (str, bool): If set, only this access_type will be looked up
                among the locks defined by `lockstring`.

        Return:
            access (bool): If check is passed or not.

        """
        try:
            if accessing_obj.locks.lock_bypass and not no_superuser_bypass:
                return True
        except AttributeError:
            if no_superuser_bypass and (
                (hasattr(accessing_obj, "is_superuser") and accessing_obj.is_superuser)
                or (
                    utils.inherits_from(accessing_obj, evennia.DefaultObject)
                    and hasattr(accessing_obj.account, "is_superuser")
                    and accessing_obj.account.is_superuser
                )
                or (
                    hasattr(accessing_obj, "get_account")
                    and (
                        not accessing_obj.get_account() or accessing_obj.get_account().is_superuser
                    )
                )
            ):
                return True
        if ":" not in lockstring:
            lockstring = "%s:%s" % ("_dummy", lockstring)

        locks = self._parse_lockstring(lockstring)

        if access_type:
            if access_type not in locks:
                return default
            else:
                return self._eval_access_type(accessing_obj, locks, access_type)
        else:
            # if no access types was given and multiple locks were
            # embedded in the lockstring we assume all must be true
            return all(
                self._eval_access_type(accessing_obj, locks, access_type) for access_type in locks
            )


# convenience access function

# dummy to be able to call check_lockstring from the outside


class _ObjDummy:
    lock_storage = ""


def check_lockstring(
    accessing_obj, lockstring, no_superuser_bypass=False, default=False, access_type=None
):
    """
    Do a direct check against a lockstring ('atype:func()..'),
    without any intermediary storage on the accessed object.

    Args:
        accessing_obj (object or None): The object seeking access.
            Importantly, this can be left unset if the lock functions
            don't access it, no updating or storage of locks are made
            against this object in this method.
        lockstring (str): Lock string to check, on the form
            `"access_type:lock_definition"` where the `access_type`
            part can potentially be set to a dummy value to just check
            a lock condition.
        no_superuser_bypass  (bool, optional): Force superusers to heed lock.
        default (bool, optional): Fallback result to use if `access_type` is set
            but no such `access_type` is found in the given `lockstring`.
        access_type (str, bool): If set, only this access_type will be looked up
            among the locks defined by `lockstring`.

    Return:
        access (bool): If check is passed or not.

    """
    global _LOCK_HANDLER
    if not _LOCK_HANDLER:
        _LOCK_HANDLER = LockHandler(_ObjDummy())
    return _LOCK_HANDLER.check_lockstring(
        accessing_obj,
        lockstring,
        no_superuser_bypass=no_superuser_bypass,
        default=default,
        access_type=access_type,
    )


def check_perm(obj, permission, no_superuser_bypass=False):
    """
    Shortcut for checking if an object has the given `permission`.  If the
    permission is in `settings.PERMISSION_HIERARCHY`, the check passes
    if the object has this permission or higher.

    This is equivalent to calling the perm() lockfunc, but without needing
    an accessed object.

    Args:
        obj (Object, Account): The object to check access. If this has a linked
            Account, the account is checked instead (same rules as per perm()).
        permission (str): The permission string to check.
        no_superuser_bypass (bool, optional): If unset, the superuser
            will always pass this check.

    """
    from evennia.locks.lockfuncs import perm

    if not no_superuser_bypass and obj.is_superuser:
        return True
    return perm(obj, None, permission)


def validate_lockstring(lockstring):
    """
    Validate so lockstring is on a valid form.

    Args:
        lockstring (str): Lockstring to validate.

    Returns:
        is_valid (bool): If the lockstring is valid or not.
        error (str or None): A string describing the error, or None
            if no error was found.

    """
    global _LOCK_HANDLER
    if not _LOCK_HANDLER:
        _LOCK_HANDLER = LockHandler(_ObjDummy())
    return _LOCK_HANDLER.validate(lockstring)


def get_all_lockfuncs():
    """
    Get a dict of available lock funcs.

    Returns:
        lockfuncs (dict): Mapping {lockfuncname:func}.

    """
    if not _LOCKFUNCS:
        _cache_lockfuncs()
    return _LOCKFUNCS


def _test():
    # testing

    class TestObj(object):
        pass

    import pdb

    obj1 = TestObj()
    obj2 = TestObj()

    # obj1.lock_storage = "owner:dbref(#4);edit:dbref(#5) or perm(Admin);examine:perm(Builder);delete:perm(Admin);get:all()"
    # obj1.lock_storage = "cmd:all();admin:id(1);listen:all();send:all()"
    obj1.lock_storage = "listen:perm(Developer)"

    pdb.set_trace()
    obj1.locks = LockHandler(obj1)
    obj2.permissions.add("Developer")
    obj2.id = 4

    # obj1.locks.add("edit:attr(test)")

    print("comparing obj2.permissions (%s) vs obj1.locks (%s)" % (obj2.permissions, obj1.locks))
    print(obj1.locks.check(obj2, "owner"))
    print(obj1.locks.check(obj2, "edit"))
    print(obj1.locks.check(obj2, "examine"))
    print(obj1.locks.check(obj2, "delete"))
    print(obj1.locks.check(obj2, "get"))
    print(obj1.locks.check(obj2, "listen"))



"""
The script handler makes sure to check through all stored scripts to
make sure they are still relevant. A scripthandler is automatically
added to all game objects. You access it through the property
`scripts` on the game object.

"""
from django.utils.translation import gettext as _

from evennia.scripts.models import ScriptDB
from evennia.utils import create, logger


class ScriptHandler(object):
    """
    Implements the handler.  This sits on each game object.

    """

    def __init__(self, obj):
        """
        Set up internal state.

        Args:
            obj (Object): A reference to the object this handler is
                attached to.

        """
        self.obj = obj

    def __str__(self):
        """
        List the scripts tied to this object.

        """
        scripts = ScriptDB.objects.get_all_scripts_on_obj(self.obj)
        string = ""
        for script in scripts:
            interval = "inf"
            next_repeat = "inf"
            repeats = "inf"
            if script.interval > 0:
                interval = script.interval
                if script.repeats:
                    repeats = script.repeats
                try:
                    next_repeat = script.time_until_next_repeat()
                except Exception:
                    next_repeat = "?"
            string += _("\n '{key}' ({next_repeat}/{interval}, {repeats} repeats): {desc}").format(
                key=script.key,
                next_repeat=next_repeat,
                interval=interval,
                repeats=repeats,
                desc=script.desc,
            )
        return string.strip()

    def add(self, scriptclass, key=None, autostart=True):
        """
        Add a script to this object.

        Args:
            scriptclass (Scriptclass, Script or str): Either a class
                object inheriting from DefaultScript, an instantiated
                script object or a python path to such a class object.
            key (str, optional): Identifier for the script (often set
                in script definition and listings)
            autostart (bool, optional): Start the script upon adding it.

        Returns:
            Script: The newly created Script.

        """
        if self.obj.__dbclass__.__name__ == "AccountDB":
            # we add to an Account, not an Object
            script = create.create_script(
                scriptclass, key=key, account=self.obj, autostart=autostart
            )
        elif isinstance(scriptclass, str) or callable(scriptclass):
            # a str or class to use create before adding to an Object. We wait to autostart
            # so we can differentiate a failing creation from a script that immediately starts/stops.
            script = create.create_script(scriptclass, key=key, obj=self.obj, autostart=False)
        else:
            # already an instantiated class
            script = scriptclass

        if not script:
            logger.log_err(f"Script {scriptclass} failed to be created.")
            return None
        if autostart:
            script.start()
        if not script.id:
            # this can happen if the script has repeats=1 or calls stop() in at_repeat.
            logger.log_info(
                f"Script {scriptclass} started and then immediately stopped; "
                "it could probably be a normal function."
            )
        return script

    def start(self, key):
        """
        Find scripts and force-start them

        Args:
            key (str): The script's key or dbref.

        Returns:
            nr_started (int): The number of started scripts found.

        """
        scripts = ScriptDB.objects.get_all_scripts_on_obj(self.obj, key=key)
        num = 0
        for script in scripts:
            script.start()
            num += 1
        return num

    def has(self, key):
        """
        Determine if a given script exists on this object.

        Args:
            key (str): Search criterion, the script's key or dbref.

        Returns:
            bool: If the script exists or not.

        """
        return ScriptDB.objects.get_all_scripts_on_obj(self.obj, key=key).exists()

    def get(self, key):
        """
        Search scripts on this object.

        Args:
            key (str): Search criterion, the script's key or dbref.

        Returns:
            scripts (queryset): The found scripts matching `key`.

        """
        return ScriptDB.objects.get_all_scripts_on_obj(self.obj, key=key)

    def remove(self, key=None):
        """
        Forcibly delete a script from this object.

        Args:
            key (str, optional): A script key or the path to a script (in the
                latter case all scripts with this path will be deleted!)
                If no key is given, delete *all* scripts on the object!

        """
        delscripts = ScriptDB.objects.get_all_scripts_on_obj(self.obj, key=key)
        if not delscripts:
            delscripts = [
                script
                for script in ScriptDB.objects.get_all_scripts_on_obj(self.obj)
                if script.path == key
            ]
        num = 0
        for script in delscripts:
            script.delete()
            num += 1
        return num

    # legacy aliases to remove
    delete = remove
    stop = delete

    def all(self):
        """
        Get all scripts stored in this handler.

        """
        return ScriptDB.objects.get_all_scripts_on_obj(self.obj)



"""
States represent the sequence of states the room goes through.

This module includes the BaseState class and the StateHandler
for managing states on the room.

The state handler operates on an Evscaperoom and changes
its state from one to another.

A given state is given as a module in states/ package. The
state is identified by its module name.

"""

from functools import wraps

from django.conf import settings

from evennia import logger, utils

from .objects import EvscaperoomObject
from .utils import create_evscaperoom_object, msg_cinematic, parse_for_things

# state setup
if hasattr(settings, "EVSCAPEROOM_STATE_PACKAGE"):
    _ROOMSTATE_PACKAGE = settings.EVSCAPEROOM_STATE_PACKAGE
else:
    _ROOMSTATE_PACKAGE = "evennia.contrib.full_systems.evscaperoom.states"
if hasattr(settings, "EVSCAPEROOM_START_STATE"):
    _FIRST_STATE = settings.EVSCAPEROOM_START_STATE
else:
    _FIRST_STATE = "state_001_start"

_GA = object.__getattribute__


# handler for managing states on room


class StateHandler(object):
    """
    This sits on the room and is used to progress through the states.

    """

    def __init__(self, room):
        self.room = room
        self.current_state_name = room.db.state or _FIRST_STATE
        self.prev_state_name = room.db.prev_state
        self.current_state = None
        self.current_state = self.load_state(self.current_state_name)

    def load_state(self, statename):
        """
        Load state without initializing it
        """
        try:
            mod = utils.mod_import(f"{_ROOMSTATE_PACKAGE}.{statename}")
        except Exception as err:
            logger.log_trace()
            self.room.msg_room(None, f"|rBUG: Could not load state {statename}: {err}!")
            self.room.msg_room(None, f"|rBUG: Falling back to {self.current_state_name}")
            return

        state = mod.State(self, self.room)
        return state

    def init_state(self):
        """
        Initialize a new state

        """
        self.current_state.init()

    def next_state(self, next_state=None):
        """
        Check if the current state is finished. This should be called whenever
        the players do actions that may affect the state of the room.

        Args:
            next_state (str, optional): If given, override the next_state given
                by the current state's check() method with this - this allows
                for branching paths (but the current state must still first agree
                that the check passes).

        Returns:
            state_changed (bool): True if the state changed, False otherwise.

        """
        # allows the state to enforce/customize what the next state should be
        next_state_name = self.current_state.next(next_state)
        if next_state_name:
            # we are ready to move on!

            next_state = self.load_state(next_state_name)
            if not next_state:
                raise RuntimeError(f"Could not load new state {next_state_name}!")

            self.prev_state_name = self.current_state_name
            self.current_state_name = next_state_name
            self.current_state.clean()
            self.prev_state = self.current_state
            self.current_state = next_state

            self.init_state()

            self.room.db.prev_state = self.prev_state_name
            self.room.db.state = self.current_state_name
            return True
        return False


# base state class


class BaseState(object):
    """
    Base object holding all callables for a state. This is here to
    allow easy overriding for child states.

    """

    next_state = "unset"
    # a sequence of hints to describe this state.
    hints = []

    def __init__(self, handler, room):
        """
        Initializer.

        Args:
            room (EvscapeRoom): The room tied to this state.
            handler (StateHandler): Back-reference to the handler
                storing this state.
        """
        self.handler = handler
        self.room = room
        # the name is derived from the name of the module
        self.name = self.__class__.__module__

    def __str__(self):
        return self.__class__.__module__

    def __repr__(self):
        return str(self)

    def _catch_errors(self, method):
        """
        Wrapper handling state method errors.

        """

        @wraps(method)
        def decorator(*args, **kwargs):
            try:
                return method(*args, **kwargs)
            except Exception:
                logger.log_trace(f"Error in State {__name__}")
                self.room.msg_room(
                    None,
                    f"|rThere was an unexpected error in State {__name__}. "
                    "Please |wreport|r this as an issue.|n",
                )
                raise  # TODO

        return decorator

    def __getattribute__(self, key):
        """
        Always wrap all callables in the error-handler

        """
        val = _GA(self, key)
        if callable(val):
            return _GA(self, "_catch_errors")(val)
        return val

    def get_hint(self):
        """
        Get a hint for how to solve this state.

        """
        hint_level = self.room.attributes.get("state_hint_level", default=-1)
        next_level = hint_level + 1
        if next_level < len(self.hints):
            # return the next hint in the sequence.
            self.room.db.state_hint_level = next_level
            self.room.db.stats["hints_used"] += 1
            self.room.log(
                f"HINT: {self.name.split('.')[-1]}, level {next_level + 1} "
                f"(total used: {self.room.db.stats['hints_used']})"
            )
            return self.hints[next_level]
        else:
            # no more hints for this state
            return None

    # helpers
    def msg(self, message, target=None, borders=False, cinematic=False):
        """
        Display messsage to everyone in room, or given target.
        """
        if cinematic:
            message = msg_cinematic(message, borders=borders)
        if target:
            options = target.attributes.get("options", category=self.room.tagcategory, default={})
            style = options.get("things_style", 2)
            # we assume this is a char
            target.msg(parse_for_things(message, things_style=style))
        else:
            self.room.msg_room(None, message)

    def cinematic(self, message, target=None):
        """
        Display a 'cinematic' sequence - centered, with borders.
        """
        self.msg(message, target=target, borders=True, cinematic=True)

    def create_object(self, typeclass=None, key="testobj", location=None, **kwargs):
        """
        This is a convenience-wrapper for quickly building EvscapeRoom objects.

        Keyword Args:
            typeclass (str): This can take just the class-name in the evscaperoom's
                objects.py module. Otherwise, a full path or the actual class
                is needed (for custom state objects, just give the class directly).
            key (str): Name of object.
            location (Object): If not given, this will be the current room.
            kwargs (any): Will be passed into create_object.
        Returns:
            new_obj (Object): The newly created object, if any.

        """
        if not location:
            location = self.room
        return create_evscaperoom_object(
            typeclass=typeclass,
            key=key,
            location=location,
            tags=[("room", self.room.tagcategory.lower())],
            **kwargs,
        )

    def get_object(self, key):
        """
        Find a named *non-character* object for this state in this room.

        Args:
            key (str): Object to search for.
        Returns:
            obj (Object): Object in the room.

        """
        match = EvscaperoomObject.objects.filter_family(
            db_key__iexact=key, db_tags__db_category=self.room.tagcategory.lower()
        )
        if not match:
            logger.log_err(f"get_object: No match for '{key}' in state ")
            return None
        return match[0]

    # state methods

    def init(self):
        """
        Initializes the state (usually by modifying the room in some way)

        """
        pass

    def clean(self):
        """
        Any cleanup operations after the state ends.

        """
        self.room.db.state_hint_level = -1

    def next(self, next_state=None):
        """
        Get the next state after this one.

        Args:
            next_state (str, optional): This allows the calling code
                to redirect to a different state than the 'default' one
                (creating branching paths in the game). Override this method
                to customize (by default the input will always override default
                set on the class)
        Returns:
            state_name (str or None): Name of next state to switch to. None
                to remain in this state. By default we check the room for the
                "finished" flag be set.
        """
        return next_state or self.next_state

    def character_enters(self, character):
        """
        Called when character enters the room in this state.

        """
        pass

    def character_leaves(self, character):
        """
        Called when character is whisked away (usually because of
        quitting). This method cannot influence the move itself; it
        happens just before room.character_cleanup()

        """
        pass
